"""Module testing the behaviour of the SubSequence class"""
import pytest

from arbok_driver import SequenceParameter
from arbok_driver.parameter_types import Time, Voltage
from .helpers import set_sweeps_args
from .conftest import (
    SquarePulse, UserSubSequence, UserSubSequenceAlt,
    square_pulse_conf, square_pulse_scalable_conf
)

nested_measurement_config = {
    'square': {
        'sequence': SquarePulse,
        'config': square_pulse_conf
    },
    'top_level_seq1': {
        'sub_sequences': {
            'nested_square': {
                'config': square_pulse_conf
            }
        }
    },
    'top_level_seq2': {
        'sub_sequences': {
            'second_level_seq': {
                'sub_sequences': {
                    'third_level_square1': {
                        'config': square_pulse_conf
                    },
                    'third_level_square2': {
                        'config': square_pulse_conf
                    },
                }
            }
        }
    }
}

def test_sub_seq_param_init(empty_sub_seq_1, empty_sub_seq_2) -> None:
    """Tests if parameters are correctly initialized from config"""
    assert empty_sub_seq_1.par1.get() == 1
    assert empty_sub_seq_2.par4.get() == 2
    assert len(empty_sub_seq_1.parameters) == 5
    assert len(empty_sub_seq_2.parameters) == 4
    assert len(empty_sub_seq_1.arbok_params) == 4
    assert len(empty_sub_seq_2.arbok_params) == 4
    assert empty_sub_seq_1.par1.get() == 1

def test_parent_child_sequence_behaviour(mock_measurement) -> None:
    """Tests the behaviour of nested sub sequences"""
    mock_measurement.add_subsequences_from_dict(nested_measurement_config)
    assert len(mock_measurement.sub_sequences) == 3
    assert len(mock_measurement.submodules) == 3
    assert len(mock_measurement.top_level_seq2.sub_sequences) == 1
    assert len(
        mock_measurement.top_level_seq2.second_level_seq.sub_sequences) == 2
    second_level_seq = mock_measurement.top_level_seq2.second_level_seq
    assert second_level_seq.third_level_square1.element.get() == "P1"
    assert len(second_level_seq.parameters) == 0
    assert second_level_seq.third_level_square1.measurement == mock_measurement

def test_add_qc_params_from_config(empty_sub_seq_1) -> None:
    """Test adding qc params by config"""
    empty_sub_seq_1.add_qc_params_from_config({
        'par8': {'type': Time, 'value': int(8)},
        'v_read': {'type': Voltage, 'elements': {'P1': 0.3, 'J1': 7, }},
        })
    assert isinstance(empty_sub_seq_1.par8, SequenceParameter)
    assert empty_sub_seq_1.par8() == 8
    assert empty_sub_seq_1.v_read_J1() == 7
    with pytest.raises(ValueError):
        empty_sub_seq_1.add_qc_params_from_config({
            'par69': {'type': Time, 'toast': int(8)},
            })

def test_find_measurement(mock_measurement) -> None:
    mock_measurement.add_subsequences_from_dict(nested_measurement_config)
    second_level_seq = mock_measurement.top_level_seq2.second_level_seq
    assert second_level_seq.measurement == mock_measurement
    assert second_level_seq.third_level_square2.measurement == mock_measurement
    assert second_level_seq.third_level_square2.parent == second_level_seq




# Single QUA script generated at 2024-06-19 16:01:00.014451
# QUA library version: 1.1.7

from qm.qua import *

with program() as prog:
    v1 = declare(int, value=0)
    with infinite_loop_():
        pause()
        align("gate_1", "gate_2", "gate_3", "gate_4")
        align("gate_1", "gate_2", "gate_3", "gate_4")
        play("unit_ramp"*amp(0.2), "gate_1", duration=20)
        play("unit_ramp"*amp(-0.1), "gate_2", duration=20)
        play("unit_ramp"*amp(0.08), "gate_3", duration=20)
        play("unit_ramp"*amp(0.25), "gate_4", duration=20)
        align("gate_1", "gate_2", "gate_3", "gate_4")
        wait(100, "gate_1", "gate_2", "gate_3", "gate_4")
        align("gate_1", "gate_2", "gate_3", "gate_4")
        play("unit_ramp"*amp(-0.2), "gate_1", duration=20)
        play("unit_ramp"*amp(0.1), "gate_2", duration=20)
        play("unit_ramp"*amp(-0.08), "gate_3", duration=20)
        play("unit_ramp"*amp(-0.25), "gate_4", duration=20)
        align("gate_1", "gate_2", "gate_3", "gate_4")
        assign(v1, (v1+1))
        r1 = declare_stream()
        save(v1, r1)
    with stream_processing():
        r1.buffer(1).save("dummy_sequence2_shots")


config = None

loaded_config = None




from .gettable_parameter import GettableParameter
from .gettable_parameter_base import GettableParameterBase
from .gettable_parameter_multi import GettableParameterMulti
from .sequence_parameter import SequenceParameter




# Single QUA script generated at 2026-03-28 03:07:59.037491
# QUA library version: 1.2.5


from qm import CompilerOptionArguments
from qm.qua import *

with program() as prog:
    v1 = declare(fixed, )
    v2 = declare(int, value=0)
    v3 = declare(int, )
    with infinite_loop_():
        pause()
        assign(v2, 0)
        assign(v3, 0)
        with while_((v3<5)):
            assign(v1, (0.1+Cast.mul_fixed_by_int(0.224775,v3)))
            align()
            align()
            play("ramp"*amp(v1), "gate_1", duration=20)
            wait(50, "gate_1")
            play("ramp"*amp(v1), "gate_1", duration=20)
            align()
            assign(v2, (v2+1))
            r0 = declare_stream()
            save(v2, r0)
            align()
            assign(v3, (v3+1))
    with stream_processing():
        r0.buffer(1).save("qm_driver_mock_measurement_shots")

config = None

loaded_config = None




from arbok_driver.parameter_types import (
    Time, Voltage, List, Int)
from arbok_driver.examples.readout_classes import (
    DcChoppedReadout
)
from arbok_driver.examples.sub_sequences import StabilityMap

stability_map_config = {
    'sequence': StabilityMap,
    'parameters': {
        'iteration': {
            'type': Int,
            'value': 0
            },
        'gate_elements': {
            'type': List,
            'value': [
                'P1', 'J1',
                'P2', 'J2',
                'P3', 'J3',
                'P4', 'J4',
                'P5', 'J5',
                'P6', 'J6',
                'P7', 'J7',
                'P8'
                ]
            },
        't_pre_chop': {
            'type': Time,
            'value': int(50e3/4)
            },
        'v_level': {
            'type': Voltage,
            'elements': {
                'P1': 0, 'J1': 0,
                'P2': 0, 'J2': 0,
                'P3': 0, 'J3': 0,
                'P4': 0, 'J4': 0,
                'P5': 0, 'J5': 0,
                'P6': 0, 'J6': 0,
                'P7': 0, 'J7': 0,
                'P8': 0,
                }
            },
        'v_home': {
            'type': Voltage,
            'elements': {
                'P1': 0, 'J1': 0,
                'P2': 0, 'J2': 0,
                'P3': 0, 'J3': 0,
                'P4': 0, 'J4': 0,
                'P5': 0, 'J5': 0,
                'P6': 0, 'J6': 0,
                'P7': 0, 'J7': 0,
                'P8': 0,
                }
            },
        },
    'signals': ['chopped_readouts'],
    'readout_groups': {
        'chop': {
            'stability': {
                'readout_class': DcChoppedReadout,
                'signal': 'chopped_readouts',
                'kwargs': {
                    'readout_qua_elements': {
                        'p1p2': 'SET1',
                        'p7p8': 'SET2',
                    }
                },
                'parameters': {
                    'v_chop': {
                        'type': Voltage,
                        'elements': {
                            'P1': 0.01,
                            'P2': -0.01,
                            'P7': 0.01,
                            'P8': -0.01,
                        }
                    },
                    'n_chops': {
                        'type': Int,
                        'value': 20
                    },
                    'chop_wait': {
                        'type': Time,
                        'value': int(20e3/4)
                    },
                },
            }
        },
    }
}




"""
Module containing a simple child class of Experiment containing a single
sub-sequence of type SquarePulse
"""
from arbok_driver import Device, Experiment
from arbok_driver.examples.configurations import square_pulse_conf

class SquarePulseExperiment(Experiment):
    """Experiment containing only a single sub-sequence of type SquarePulse"""
    _name = 'square_pulse_experiment'

    @property
    def sequences_config(self):
        """Sequences-dict to contruct experiment"""
        return {
            'square_pulse': {'config': square_pulse_conf}
        }

class SquarePulseExperimentWithDefaults(Experiment):
    """
    Experiment containing only a single sub-sequence of type SquarePulse
    that can be given as argument or can be loaded from the measurements defaults
    """
    _name = 'square_pulse_experiment'
    def __init__(
        self,
        square_pulse_config: dict | None = None
    ):
        super().__init__(square_pulse =  square_pulse_config)

    @property
    def sequences_config(self):
        """Sequences-dict to contruct experiment"""
        return {
            'square_pulse': self.configs['square_pulse']
        }



""" Module containing ParameterType classes """
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Generic, TypeVar

import numpy as np
from qm import qua
from qm.qua._expressions import QuaVariable, Scalar
from qcodes.validators import (
    Arrays, Numbers, Ints, MultiTypeOr, Strings, Sequence, Bool
)
from .parameters.sequence_parameter import SequenceParameter

K = TypeVar("K", bound=str)
V = TypeVar("V", covariant = True)

@dataclass(frozen=True)
class ParameterMap(Mapping[K, V], Generic[K, V]):
    _mapping: Mapping[K, V]

    def __getitem__(self, key: K) -> V:
        return self._mapping[key]

    def __iter__(self):
        return iter(self._mapping)

    def __len__(self) -> int:
        return len(self._mapping)

class Time(SequenceParameter[int]):
    """
    Cycles sequence parameter. Parameter values and sweeps are given in units of 
    FPGA cycles (4 ns). The parameter is converted to seconds before being
    registered to a measurement.
    """
    unit = 's'
    """ Default: 'cycles' """
    var_type = int
    """ Default: int """
    vals = Ints()
    scale = 1
    sweep_validator = None
    ### TODO: Implement vals. User should only be able to set ints (cycles)
    """ Default: Numbers """

    def convert_to_real_units(self, value):
        return np.array(np.array(value)*4e-9)

class String(SequenceParameter[str]):
    unit = 'N/A'
    """ Default: 'N/A' """
    var_type = str
    """ Default: str """
    vals = Strings()
    scale = None
    """ Default: Strings """
    sweep_validator = MultiTypeOr(Strings(), Arrays(valid_types = [None]))

class Voltage(SequenceParameter[float | QuaVariable]):
    unit = 'V'
    """ Default: 'V' """
    var_type = qua.fixed
    """ Default: fixed """
    vals = Numbers()
    scale = 1
    sweep_validator = MultiTypeOr(Numbers(), Arrays(valid_types = [float]))
    """ Default: Numbers """

class Frequency(SequenceParameter[int | QuaVariable]):
    unit = 'Hz'
    """ Default: 'Hz' """
    var_type = int
    """ Default: 'qua.fixed' """
    vals = Numbers()
    scale = 1
    sweep_validator = MultiTypeOr(Numbers(), Arrays(valid_types = [int]))
    """ Default: Numbers """

class Amplitude(SequenceParameter[float | Scalar]):
    unit = None
    var_type = qua.fixed
    """ Default: fixed """
    vals = Numbers(min_value = -2, max_value = 2)
    scale = 1
    sweep_validator = MultiTypeOr(
        Numbers(min_value = -2, max_value = 2),
        Arrays(valid_types = [float])
        )
    """ Default: Numbers """

class List(SequenceParameter[list]):
    unit = 'N/A'
    """ Default: 'N/A' """
    var_type = None
    """ Default: str """
    vals = Sequence()
    scale = None
    """ Default: Strings """
    sweep_validator = Sequence()

class Int(SequenceParameter[int | QuaVariable]):
    var_type = int
    """ Default: int """
    vals = Ints()
    scale = 1
    sweep_validator = MultiTypeOr(Numbers(), Arrays(valid_types = [int]))

class Boolean(SequenceParameter[bool]):
    var_type = bool
    """ Default: int """
    vals = Bool()
    scale = None
    sweep_validator = None

class Radian(SequenceParameter[float]):
    unit = 'pi'
    var_type = qua.fixed
    vals = Numbers(min_value = -2*np.pi, max_value = 2*np.pi)
    scale = 1
    sweep_validator = MultiTypeOr(
        Numbers(min_value = -2*np.pi, max_value = 2*np.pi),
        Arrays(valid_types = [float])
        )

class Pi(SequenceParameter):
    unit = 'pi'
    var_type = qua.fixed
    scale = 1
    vals = Numbers(min_value = -2, max_value = 2)
    sweep_validator = MultiTypeOr(
        Numbers(min_value = -2, max_value = 2),
        Arrays(valid_types = [float])
        )



"""Module containing various utils"""
import logging
from importlib.util import spec_from_file_location, module_from_spec

from anytree import Node
import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from qcodes.instrument import Instrument
from qcodes.station import Station

def plot_qmm_simulation_results(simulated_devices):
    """ 
    Visualizes analog and digital channel simulation results 
    TODO: Check type of `simulated_devices`
    """
    fig, [a, b] = plt.subplots(2, sharex= True)
    for channel, data in simulated_devices.analog.items():
        a.plot(data, label = channel)
    for channel, data in simulated_devices.digital.items():
        b.plot(data, label = channel)

    ncols_a = int((len(a.lines)-1)/10) + 1
    a.legend(bbox_to_anchor=(1.0, 1.0), loc='upper left', fontsize = 8,
                title = 'analog', ncols = ncols_a)
    a.grid()
    a.set_ylabel("Voltage in V")
    a.set_title("simulated analog/digital outputs", loc = 'right')

    ncols_b = int((len(b.lines)-1)/10) + 1
    b.legend(bbox_to_anchor=(1.0, 1.0), loc='upper left', fontsize = 8,
                title = 'digital', ncols = ncols_b)
    b.grid()
    b.set_xlabel("Time in ns")
    b.set_ylabel("Digital Signal")
    fig.subplots_adjust(wspace=0, hspace=0)

def plotly_sim_results(simulated_devices):
    """
    Plots one graph for each analog and digital simulation result per controller
    
    Args:
        simulated_results: simulated devices from qm simulator
        
    Returns:
        plotly figure
    """
    controller_dict = get_all_controller_results(simulated_devices)
    fig = make_subplots(rows=1, cols=1)
    for index, results in controller_dict.items():
        for i in range(1, 11):
            if str(i) in results.analog:
                fig.add_trace(
                    go.Scatter(
                        x = list(range(len(results.analog[str(i)]))),
                        y = results.analog[str(i)],
                        mode='lines',
                        name = f'con{index} analog ch{i}',
                        opacity=0.8,
                        visible = True
                        ),
                    )
        for i in range(1, 11):
            if str(i) in results.digital:
                fig.add_trace(
                    go.Scatter(
                        x = list(range(len(results.digital[str(i)]))),
                        y = np.array(results.digital[str(i)], dtype = int),
                        mode='lines',
                        name = f'con{index} digital ch{i}',
                        opacity=0.8,
                        visible='legendonly',
                        ),
                    )
        fig.update_xaxes(title_text="time in ns", row=1, col=1)
        fig.update_yaxes(title_text="voltage in V", row=1, col=1)

    fig.update_layout(
        autosize=True,
        margin=dict(l=50, r=50, t=50, b=50),
        title_text="Simulation results on all available quantum machines",
        legend_tracegroupgap = 100,
    )
    return fig

def plotly_sim_results_separate(simulated_devices):
    """
    Plots one graph for each analog and digital simulation result per controller
    
    Args:
        simulated_results: simulated devices from qm simulator
        
    Returns:
        plotly figure
    """
    controller_dict = get_all_controller_results(simulated_devices)
    subplot_titles = []
    for index, _ in controller_dict.items():
        subplot_titles.append(f"Con{index}: analog")
        subplot_titles.append(f"Con{index}: digital")
    fig = make_subplots(
        rows=2*len(controller_dict),
        cols=1,
        subplot_titles=subplot_titles,
        vertical_spacing = 0.1,
    )
    plot_counter = 0
    for index, results in controller_dict.items():
        plot_counter +=1 
        for i in range(1, 11):
            row = (index-1)*2+1
            if str(i) in results.analog:
                fig.add_trace(
                    go.Scatter(
                        x = list(range(len(results.analog[str(i)]))),
                        y = results.analog[str(i)],
                        mode='lines',
                        name = f'con{index} analog ch{i}',
                        opacity=0.8,
                        #legendgroup= str(row),
                        visible = True
                        ),
                    row=(index-1)*2+1,
                    col=1
                    )
        fig.update_xaxes(title_text="time in ns", row=row, col=1)
        fig.update_yaxes(title_text="voltage in V", row=row, col=1)
        for i in range(1, 11):
            if str(i) in results.digital:
                row = index*2
                plot_counter +=1
                fig.add_trace(
                    go.Scatter(
                        x = list(range(len(results.digital[str(i)]))),
                        y = np.array(results.digital[str(i)], dtype = int),
                        mode='lines',
                        name = f'con{index} digital ch{i}',
                        opacity=0.8,
                        visible=True,
                        #legendgroup= str(row),
                        ),
                    row=index*2,
                    #row=(index-1)*2+1,
                    col=1,
                    )
        fig.update_xaxes(title_text="time in ns", row=row, col=1)
        fig.update_yaxes(title_text="voltage in V", row=row, col=1)

    fig.update_layout(
        autosize=False,
        width=800,
        height=plot_counter*300,
        margin=dict(l=50, r=50, t=50, b=50),
        title_text="Simulation results on all available quantum machines",
        legend_tracegroupgap = 100,
    )
    return fig

def get_all_controller_results(simulated_devices: any) -> dict:
    """Takes simulated devices and returns a dict with all controller results"""
    controller_dict = {}

    for name, attr in simulated_devices.__dict__.items():
        if 'con' in name:
            controller_dict[name.split('con')[1]] = attr
    if len(controller_dict) == 0:
        raise ValueError("No controller results found!")
    return controller_dict

def set_qm_logging_to_file(filename):
    """Changes the standart logging handle (stdout) to the given file"""
    logger = logging.getLogger('qm')
    logger.setLevel('DEBUG')

    qm_log_file = open(filename, 'w')
    lh = logging.StreamHandler(qm_log_file)
    logger.addHandler(lh)
    for handler in logger.handlers:
        logger.removeHandler(handler)

def get_module(name: str = None, mod_path: str = None):
    """
    Dynamically loads a Python module from a specified file path.

    This function creates a module specification and loads the module from the
    provided file path.

    Args:
        name: The desired name for the loaded module.
        mod_path: The absolute path to the Python module file (.py).

    Returns:
        The loaded module object.

    Raises:
        ImportError: If the module specification cannot be created or if the
                     module cannot be loaded.
    """
    spec = spec_from_file_location(name, mod_path)
    if spec is None:
        raise ImportError(f"Cannot create module spec for {mod_path}.")
    mod = module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod

def dict_to_anytree(name, d, parent=None):
    """Convert a nested dictionary to an anytree structure."""
    node = Node(name, parent=parent)
    for key, val in d.items():
        dict_to_anytree(key, val, parent=node)
    return node



"""Module containing difference readout helper"""
from qm import qua
from arbok_driver import (
    AbstractReadout,
    ReadSequence,
    Signal, 
)

class Difference(AbstractReadout):
    """Helper class enabling difference readout from a given signal"""

    def __init__(
            self,
            name: str,
            read_sequence: ReadSequence,
            signal: Signal,
            save_results: bool,
            parameters: dict,
            minuend: str,
            subtrahend: str,
            ):
        """
        Constructor method of `Difference` readout helper
        
        Args:
            name (str): Name of readout
            read_sequence (arbok_driver.ReadSequence): ReadSequence performing the
                abstract readout
            signal (arbok_driver.Signal): Signal to be measured
            save_results (bool): Whether to save the results
            params (dict): Parameters to be added to the read sequence
            minuend (str): Name of the minuend gettable
            subtrahend (str): Name of the subtrahend gettable
        """
        super().__init__(
            name = name,
            read_sequence = read_sequence,
            signal = signal,
            save_results = save_results,
            parameters = parameters
        )
        self.signal = signal
        self.minuend = minuend
        self.subtrahend = subtrahend
        self._create_gettables()

    def qua_measure(self):
        """
        Measures the given gettables and assigns the result to the variables.
        Only to be called within qua.program() context manager!
        """
        qua.assign(
            self.difference_gettable.qua_result_var,
            self.minuend_gettable.qua_result_var
            - self.subtrahend_gettable.qua_result_var
            )

    def _create_gettables(self):
        """Creates the gettables for the given readout"""
        self.minuend_gettable = self.get_gettable_from_path(self.minuend)
        self.subtrahend_gettable = self.get_gettable_from_path(self.subtrahend)

        self.difference_gettable = self.create_gettable(
            gettable_name = self.name,
            var_type = qua.fixed
        )



"""
This file contains the quantum machine configuration for the nova sample.
"""
import os
import numpy as np

def round_to_multiple(x, base=4):
    return int(base * round(float(x)/base))

### CONSTANTS
UNIT_AMP = 0.4999
UNIT_AMP_AMPLIFIED = 2.4999

UNIT_RAMP_PULSE_DURATION = int(20)
MEASUREMENT_PULSE_DURATION = int(50e3) # 100e3
NORMALIZATION_FACTOR = MEASUREMENT_PULSE_DURATION/100e3
CONST_INTEGRATION_WEIGHT = 0.1/NORMALIZATION_FACTOR

ESR_CHIRP_LENGTH = int(100e3) # 100e3
MIN_PULSE_LENGTH =int(100)

path = os.path.dirname(os.path.abspath(__file__)) + '/'

LO_Q1 = int(13.959e9)
LO_Q2 = int(13.959e9)
LO_Q3 = int(13.959e9)
LO_Q4 = int(13.959e9)
LO_Q5 = int(13.959e9)
LO_Q6 = int(13.959e9)
LO_Q7 = int(13.959e9)
LO_Q8 = int(13.959e9)

IQampQ1 = 0.2
IQamp_max = 0.2

### CONTROLLERS
con1 = 'con1'

mw_fem = 1
lf_fem1 = 2
lf_fem2 = 4
lf_fem3 = 5
lf_fem4 = 7

sampling_rate = int(1e9)


import random

qubits = [f"Q{i}" for i in range(1, 13)]

# Larmor frequencies: 1e6 to 100e6 (ints)
larmor_freqs = {
    q: random.randint(int(1e6), int(100e6))
    for q in qubits
}

# Sideband (keeping your single entry)
sideband_freqs = {
    "qe1": random.randint(int(1e6), int(100e6))
}

# 2π times: 1e2 to 1e3 (floats or ints depending on your needs)
t_2pi_times = {  # in ns
    q: random.uniform(1e2, 1e3)
    for q in qubits
}

# Amplitudes: 0.8 to 1.0
amp_x_gate = {
    q: random.uniform(0.8, 1.0)
    for q in qubits
}

opx1000_config = {
    #'version': 1,
    'controllers': {
        con1: {
            "type": "opx1000",
            "fems": {
                mw_fem: {
                    "type": "MW",
                    "analog_outputs": {
                        2: {
                                #"sampling_rate": sampling_rate,
                                "band": 3,
                                "full_scale_power_dbm": -10,
                                "upconverters": {
                                    1: {"frequency": int(8.4e9)},
                                },
                            },
                    },
                    "digital_outputs": {},
                },
                lf_fem1: {
                    "type": "LF",
                    "analog_outputs": {
                        1: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        2: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        3: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        4: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        5: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        6: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        7: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        8: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                    },
                    "digital_outputs": {
                        1: {},
                        2: {},
                        3: {},
                        4: {},
                        5: {},
                        6: {},
                        7: {},
                        8: {},
                    },
                    'analog_inputs': {
                        1: {'offset': 0},  # nothing
                        2: {'offset': 0},  # nothing
                    },
                },
                lf_fem2: {
                    "type": "LF",
                    "analog_outputs": {
                        1: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        2: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        3: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        4: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        5: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        6: {
                            "offset": 0.5,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        7: {
                            "offset": 0.5,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        8: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                    },
                    "digital_outputs": {
                        1: {},
                        2: {},
                        3: {},
                        4: {},
                        5: {},
                        6: {},
                        7: {},
                        8: {},
                    },
                    'analog_inputs': {
                        1: {'offset': 0},  # nothing
                        2: {'offset': 0},  # nothing
                    },
                },
                lf_fem3: {
                    "type": "LF",
                    "analog_outputs": {
                        1: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        2: {
                            "offset": 0.4,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        3: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        4: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        5: {
                            "offset": 0.0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        6: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        7: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        8: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                    },
                    "digital_outputs": {
                        1: {},
                        2: {},
                        3: {},
                        4: {},
                        5: {},
                        6: {},
                        7: {},
                        8: {},
                    },
                    'analog_inputs': {
                        1: {'offset': 0},  # nothing
                        2: {'offset': 0},  # nothing
                    },
                },
                lf_fem4: {
                    "type": "LF",
                    "analog_outputs": {
                        1: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        2: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        3: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        4: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        5: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        6: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        7: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                        8: {
                            "offset": 0,
                            "output_mode": "amplified",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "pulse",
                        },
                    },
                    "digital_outputs": {
                        1: {},
                        2: {},
                        3: {},
                        4: {},
                        5: {},
                        6: {},
                        7: {},
                        8: {},
                    },
                    'analog_inputs': {
                        1: {'offset': 0},  # nothing
                        2: {'offset': 0},  # nothing
                    },
                },
            },
        }
    },
    'elements': {
        'P1': {
            'singleInput': {'port': (con1, lf_fem3, 1)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P2': {
            'singleInput': {'port': (con1, lf_fem4, 4)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P3': {
            'singleInput': {'port': (con1, lf_fem4, 3)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P4': {
            'singleInput': {'port': (con1, lf_fem4, 2)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P5': {
            'singleInput': {'port': (con1, lf_fem4, 1)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P6': {
            'singleInput': {'port': (con1, lf_fem1, 5)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P7': {
            'singleInput': {'port': (con1, lf_fem1, 6)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P8': {
            'singleInput': {'port': (con1, lf_fem1, 7)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P9': {
            'singleInput': {'port': (con1, lf_fem1, 8)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P10': {
            'singleInput': {'port': (con1, lf_fem4, 2)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P11': {
            'singleInput': {'port': (con1, lf_fem2, 6)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'P12': {
            'singleInput': {'port': (con1, lf_fem2, 7)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },



       'J1': {
        'singleInput': {'port': (con1, lf_fem1, 1)},
        'sticky': {
            'analog': True, 
            'digital': False, 
            'duration': 100
        },
        'operations': {
            'unit_ramp': 'unit_ramp_pulse_2_5',
        },
        },

        'J2': {
            'singleInput': {'port': (con1, lf_fem1, 2)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'J3': {
            'singleInput': {'port': (con1, lf_fem1, 3)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'J4': {
            'singleInput': {'port': (con1, lf_fem1, 4)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'J5': {
            'singleInput': {'port': (con1, lf_fem2, 1)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'J6': {
            'singleInput': {'port': (con1, lf_fem2, 2)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'J7': {
            'singleInput': {'port': (con1, lf_fem2, 3)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'J8': {
            'singleInput': {'port': (con1, lf_fem3, 7)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'J9': {
            'singleInput': {'port': (con1, lf_fem3, 4)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'J10': {
            'singleInput': {'port': (con1, lf_fem3, 3)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'J11': {
            'singleInput': {'port': (con1, lf_fem3, 2)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
            },
        },

        'SET1': {
            'singleInput': {
                'port': (con1, lf_fem2, 8)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 400
            },
            'outputs': {
                'output1': (con1, lf_fem2, 1)
            },
            'digitalInputs':{
                'digital_input1':{
                    'port': (con1, lf_fem2, 1),
                    'delay': 24,
                    'buffer': 0
                    }
               },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
                'step':'stepPulse',
                'measure': 'measure_pulse',
                'measure_1us': 'measure_pulse_1us',
            },
            'time_of_flight': 180,
            'smearing': 0,
            'intermediate_frequency': 0,
        },

        'SET2': {
            'singleInput': {
                'port': (con1, lf_fem3, 8)},
            'sticky':{
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'outputs': {
                'output1': (con1, lf_fem3, 1)
            },
            'digitalInputs':{
                'digital_input1':{
                    'port': (con1, lf_fem3, 1),
                    'delay': 24,
                    'buffer': 0
                    }
               },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
                'step':'stepPulse',
                'measure': 'measure_pulse',
                'measure_1us': 'measure_pulse_1us',
            },
            'time_of_flight': 180,
            'smearing': 0,
            'intermediate_frequency': 0,
        },

        'SET3': {
            'singleInput': {'port': (con1, lf_fem2, 4)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'outputs': {
                'output1': (con1, lf_fem2, 2)
            },
            'digitalInputs':{
                'digital_input1':{
                    'port': (con1, lf_fem2, 2),
                    'delay': 24,
                    'buffer': 0
                    }
               },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
                'step':'stepPulse',
                'measure': 'measure_pulse',
                'measure_1us': 'measure_pulse_1us',
                'measure_1ms': 'measure_pulse_1ms',
            },
            'time_of_flight': 180,
            'smearing': 0,
            'intermediate_frequency': 0,
        },

        'SET4': {
            'singleInput': {'port': (con1, lf_fem3, 6)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'outputs': {
                'output1': (con1, lf_fem3, 2)
            },
            'digitalInputs':{
                'digital_input1':{
                    'port': (con1, lf_fem3, 2),
                    'delay': 24,
                    'buffer': 0
                    }
               },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
                'step':'stepPulse',
                'measure': 'measure_pulse',
                'measure_1us': 'measure_pulse_1us',
                'measure_1ms': 'measure_pulse_1ms',
            },
            'time_of_flight': 180,
            'smearing': 0,
            'intermediate_frequency': 0,
        },

        'SET5': {
            'singleInput': {'port': (con1, lf_fem4, 5)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'outputs': {
                'output1': (con1, lf_fem4, 1)
            },
            'digitalInputs':{
                'digital_input1':{
                    'port': (con1, lf_fem4, 1),
                    'delay': 24,
                    'buffer': 0
                    }
               },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
                'step':'stepPulse',
                'measure': 'measure_pulse',
                'measure_1us': 'measure_pulse_1us',
                'measure_1ms': 'measure_pulse_1ms',
            },
            'time_of_flight': 180,
            'smearing': 0,
            'intermediate_frequency': 0,
        },

        'SET6': {
            'singleInput': {'port': (con1, lf_fem4, 8)},
            'sticky': {
                'analog': True, 
                'digital': False, 
                'duration': 100
            },
            'outputs': {
                'output1': (con1, lf_fem4, 2)
            },
            'digitalInputs':{
                'digital_input1':{
                    'port': (con1, lf_fem4, 2),
                    'delay': 24,
                    'buffer': 0
                    }
               },
            'operations': {
                'unit_ramp': 'unit_ramp_pulse_2_5',
                'step':'stepPulse',
                'measure': 'measure_pulse',
                'measure_1us': 'measure_pulse_1us',
                'measure_1ms': 'measure_pulse_1ms',
            },
            'time_of_flight': 180,
            'smearing': 0,
            'intermediate_frequency': 0,
        },

        "qe1": {
            'mixInputs': {
                'I': (con1, lf_fem3, 5),
                'Q': (con1, lf_fem3, 7),
                'lo_frequency': LO_Q1,
                'mixer': 'mixer_qe1'
            },
            'digitalInputs': {
                "switch1": {
                    "port": (con1, lf_fem3, 1),
                    "delay": 76,
                    "buffer": 4,
                },
            },
            'intermediate_frequency': sideband_freqs['qe1'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q1_pi_pulse',
                'control_pi2': 'ESR_Q1_pi2_pulse',
                'control_minus_pi2': 'ESR_Q1_pi2_pulse_minus',

            },
        },

        'Q1': {
            # MWInput corresponds to an OPX physical output port
            "MWInput": {
                "port": (con1, mw_fem, 2),
                "upconverter": 1,
            },
            'intermediate_frequency': larmor_freqs['Q1'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q1_pi_pulse',
                'control_pi2': 'ESR_Q1_pi2_pulse',
                'control_minus_pi2': 'ESR_Q1_pi2_pulse_minus',

            },
        },
        'Y1': {
            "MWInput": {
                "port": (con1, mw_fem, 2),
                "upconverter": 1,
            },
            'intermediate_frequency': larmor_freqs['Q1'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q1_pi_pulse',
                'control_pi2': 'ESR_Q1_pi2_pulse',
                'control_minus_pi2': 'ESR_Q1_pi2_pulse_minus',

            },
        },
        'Q2': {
            "MWInput": {
                "port": (con1, mw_fem, 2),
                "upconverter": 1,
            },
            'intermediate_frequency': larmor_freqs['Q2'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q2_pi_pulse',
                'control_pi2': 'ESR_Q2_pi2_pulse',
                'control_minus_pi2': 'ESR_Q2_pi2_pulse_minus',

            },
        },
        'Y2': {
            "MWInput": {
                "port": (con1, mw_fem, 2),
                "upconverter": 1,
            },
            'intermediate_frequency': larmor_freqs['Q2'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q2_pi_pulse',
                'control_pi2': 'ESR_Q2_pi2_pulse',
                'control_minus_pi2': 'ESR_Q2_pi2_pulse_minus',

            },
        },
        'Q3': {
            "MWInput": {
                "port": (con1, mw_fem, 2),
                "upconverter": 1,
            },
            'intermediate_frequency': larmor_freqs['Q3'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q3_pi_pulse',
                'control_pi2': 'ESR_Q3_pi2_pulse',
                'control_minus_pi2': 'ESR_Q3_pi2_pulse_minus',

            },
        },
        'Q4': {
            "MWInput": {
                "port": (con1, mw_fem, 2),
                "upconverter": 1,
            },
            'intermediate_frequency': larmor_freqs['Q4'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q4_pi_pulse',
                'control_pi2': 'ESR_Q4_pi2_pulse',
                'control_minus_pi2': 'ESR_Q4_pi2_pulse_minus',

            },
        },
        'Q5': {
            "MWInput": {
                "port": (con1, mw_fem, 2),
                "upconverter": 1,
            },
            'intermediate_frequency': larmor_freqs['Q5'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q5_pi_pulse',
                'control_pi2': 'ESR_Q5_pi2_pulse',
                'control_minus_pi2': 'ESR_Q5_pi2_pulse_minus',

            },
        },
        'Q6': {
            "MWInput": {
                "port": (con1, mw_fem, 2),
                "upconverter": 1,
            },
            'intermediate_frequency': larmor_freqs['Q6'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q6_pi_pulse',
                'control_pi2': 'ESR_Q6_pi2_pulse',
                'control_minus_pi2': 'ESR_Q6_pi2_pulse_minus',

            },
        },
        'Q7': {
            "MWInput": {
                "port": (con1, mw_fem, 2),
                "upconverter": 1,
            },
            'intermediate_frequency': larmor_freqs['Q7'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q7_pi_pulse',
                'control_pi2': 'ESR_Q7_pi2_pulse',
                'control_minus_pi2': 'ESR_Q7_pi2_pulse_minus',

            },
        },
        'Q8': {
            "MWInput": {
                "port": (con1, mw_fem, 2),
                "upconverter": 1,
            },
            'intermediate_frequency': larmor_freqs['Q8'],
            'operations': {
                'control': 'ESR_const_pulse',
                'chirp': 'ESR_chirp_pulse',
                'control_pi': 'ESR_Q8_pi_pulse',
                'control_pi2': 'ESR_Q8_pi2_pulse',
                'control_minus_pi2': 'ESR_Q8_pi2_pulse_minus',

            },
        },
    },

    'pulses': {
        'trigger_pulse':{
            'operation': 'measurement',
            'length':1000,
            'digital_marker': 'ON',
        },

        'measure_pulse': {
            'operation': 'measurement',
            'length': MEASUREMENT_PULSE_DURATION,
            'waveforms': {
                    'single': 'zero_wf'
                },
            'digital_marker': 'ON',
            'integration_weights': {
                'x': 'cos',
                'y': 'sin',
                'x_const': 'x_const',
                'y_const': 'y_const',
                'x_test': 'cos_test',
                'y_test': 'sin_test'
                }
        },
        'measure_pulse_1us': {
            'operation': 'measurement',
            'length': int(1e3),
            'waveforms': {
                    'single': 'zero_wf'
                },
            'digital_marker': 'ON',
            'integration_weights': {
                'x': 'cos',
                'y': 'sin',
                'x_const': 'x_const_1us',
                'y_const': 'y_const',
                'x_test': 'cos_test',
                'y_test': 'sin_test'
                }
        },
        'measure_pulse_1ms': {
            'operation': 'measurement',
            'length': int(1e6),
            'waveforms': {
                    'single': 'zero_wf'
                },
            'digital_marker': 'ON',
            'integration_weights': {
                'x_const_1ms': 'x_const_1ms',
                'y_const_1ms': 'y_const_1ms',
                }
        },
         'unit_ramp_pulse': {
            'operation': 'control',
            'length': UNIT_RAMP_PULSE_DURATION,
            'waveforms': {
                'single': 'unit_ramp_wf',
            }
        },
         'unit_ramp_pulse_2_5': {
            'operation': 'control',
            'length': UNIT_RAMP_PULSE_DURATION,
            'waveforms': {
                'single': 'unit_ramp_2_5_wf',
            }
        },
        "stepPulse": {
            "operation": "control",
            "length": 1000,
            "waveforms": {"single": "step_wf"},
        },
        "const_pulse": {
            "operation": "control",
            "length": int(10000),
            "waveforms": {
                "I": "const_wf",
                "Q": "zero_wf",
            },
        },
        'ESR_const_pulse': {
            'operation': 'control',
            'length': MIN_PULSE_LENGTH,
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf',
                'Q': 'max_amp_wf'
            }
        },
        'ESR_chirp_pulse': {
            'operation': 'control',
            'length': ESR_CHIRP_LENGTH,
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'q1_ctrl_wf',
                'Q': 'q1_ctrl_wf'
            }
        },
        'ESR_Q1_pi_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q1']/2),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q1',
                'Q': 'max_amp_wf_Q1'
            }
        },
        'ESR_Q1_pi2_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q1']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q1',
                'Q': 'max_amp_wf_Q1'
            }
        },
        'ESR_Q1_pi2_pulse_minus': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q1']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q1_minus',
                'Q': 'max_amp_wf_Q1_minus'
            }
        },
        'ESR_Q2_pi_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q2']/2),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q2',
                'Q': 'max_amp_wf_Q2'
            }
        },
        'ESR_Q2_pi2_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q2']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q2',
                'Q': 'max_amp_wf_Q2'
            }
        },
        'ESR_Q2_pi2_pulse_minus': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q2']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q2_minus',
                'Q': 'max_amp_wf_Q2_minus'
            }
        },
        'ESR_Q3_pi_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q3']/2),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q3',
                'Q': 'max_amp_wf_Q3'
            }
        },
        'ESR_Q3_pi2_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q3']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q3',
                'Q': 'max_amp_wf_Q3'
            }
        },
        'ESR_Q3_pi2_pulse_minus': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q3']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q3_minus',
                'Q': 'max_amp_wf_Q3_minus'
            }
        },
        'ESR_Q4_pi_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q4']/2),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q4',
                'Q': 'max_amp_wf_Q4'
            }
        },
        'ESR_Q4_pi2_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q4']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q4',
                'Q': 'max_amp_wf_Q4'
            }
        },
        'ESR_Q4_pi2_pulse_minus': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q4']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q4_minus',
                'Q': 'max_amp_wf_Q4_minus'
            }
        },
        'ESR_Q5_pi_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q5']/2),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q5',
                'Q': 'max_amp_wf_Q5'
            }
        },
        'ESR_Q5_pi2_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q5']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q5',
                'Q': 'max_amp_wf_Q5'
            }
        },
        'ESR_Q5_pi2_pulse_minus': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q5']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q5_minus',
                'Q': 'max_amp_wf_Q5_minus'
            }
        },
        'ESR_Q6_pi_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q6']/2),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q6',
                'Q': 'max_amp_wf_Q6'
            }
        },
        'ESR_Q6_pi2_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q6']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q6',
                'Q': 'max_amp_wf_Q6'
            }
        },
        'ESR_Q6_pi2_pulse_minus': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q6']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q6_minus',
                'Q': 'max_amp_wf_Q6_minus'
            }
        },
        'ESR_Q7_pi_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q7']/2),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q7',
                'Q': 'max_amp_wf_Q7'
            }
        },
        'ESR_Q7_pi2_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q7']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q7',
                'Q': 'max_amp_wf_Q7'
            }
        },
        'ESR_Q7_pi2_pulse_minus': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q7']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q7_minus',
                'Q': 'max_amp_wf_Q7_minus'
            }
        },
        'ESR_Q8_pi_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q8']/2),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q8',
                'Q': 'max_amp_wf_Q8'
            }
        },
        'ESR_Q8_pi2_pulse': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q8']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q8',
                'Q': 'max_amp_wf_Q8'
            }
        },
        'ESR_Q8_pi2_pulse_minus': {
            'operation': 'control',
            'length': round_to_multiple(t_2pi_times['Q8']/4),
            'digital_marker': 'ON',
            'waveforms': {
                'I': 'max_amp_wf_Q8_minus',
                'Q': 'max_amp_wf_Q8_minus'
            }
        },
        'constPulse': {
            'operation': 'control',
            'length': 10000,  # in ns
            'waveforms': {
                'I': 'const_wf',
                'Q': 'zero_wf',
            },
            'digital_marker': 'ON',
        },
    },

    'waveforms': {
        'zero_wf': {
            'type': 'constant',
            'sample': 0.0
        },
        "const_wf": {
            "type": "constant",
            "sample": 0.2
        },
        "step_wf": {
            'type': 'constant',
            'sample': UNIT_AMP
            },
        'unit_ramp_wf':{
            'type':'arbitrary',
            'samples': np.linspace(0, UNIT_AMP, UNIT_RAMP_PULSE_DURATION).tolist()
            },
        'unit_ramp_2_5_wf':{
            'type':'arbitrary',
            'samples': np.linspace(0, UNIT_AMP_AMPLIFIED, UNIT_RAMP_PULSE_DURATION).tolist()
            },
        'q1_ctrl_wf': {
            'type': 'constant',
            'sample': IQampQ1
        },
        'max_amp_wf': {
            'type': 'constant',
            'sample': IQamp_max
        },
        'max_amp_wf_Q1': {
            'type': 'constant',
            'sample': IQamp_max*amp_x_gate['Q1']
        },
        'max_amp_wf_Q1_minus': {
            'type': 'constant',
            'sample': -IQamp_max*amp_x_gate['Q1']
        },
        'max_amp_wf_Q2': {
            'type': 'constant',
            'sample': IQamp_max*amp_x_gate['Q2']
        },
        'max_amp_wf_Q2_minus': {
            'type': 'constant',
            'sample': -IQamp_max*amp_x_gate['Q2']
        },
        'max_amp_wf_Q3': {
            'type': 'constant',
            'sample': IQamp_max*amp_x_gate['Q3']
        },
        'max_amp_wf_Q3_minus': {
            'type': 'constant',
            'sample': -IQamp_max*amp_x_gate['Q3']
        },
        'max_amp_wf_Q4': {
            'type': 'constant',
            'sample': IQamp_max*amp_x_gate['Q4']
        },
        'max_amp_wf_Q4_minus': {
            'type': 'constant',
            'sample': -IQamp_max*amp_x_gate['Q4']
        },
        'max_amp_wf_Q5': {
            'type': 'constant',
            'sample': IQamp_max*amp_x_gate['Q5']
        },
        'max_amp_wf_Q5_minus': {
            'type': 'constant',
            'sample': -IQamp_max*amp_x_gate['Q5']
        },
        'max_amp_wf_Q6': {
            'type': 'constant',
            'sample': IQamp_max*amp_x_gate['Q6']
        },
        'max_amp_wf_Q6_minus': {
            'type': 'constant',
            'sample': -IQamp_max*amp_x_gate['Q6']
        },
        'max_amp_wf_Q7': {
            'type': 'constant',
            'sample': IQamp_max*amp_x_gate['Q7']
        },
        'max_amp_wf_Q7_minus': {
            'type': 'constant',
            'sample': -IQamp_max*amp_x_gate['Q7']
        },
        'max_amp_wf_Q8': {
            'type': 'constant',
            'sample': IQamp_max*amp_x_gate['Q8']
        },
        'max_amp_wf_Q8_minus': {
            'type': 'constant',
            'sample': -IQamp_max*amp_x_gate['Q8']
        },
    },

    'digital_waveforms': {
        'ON': {
            'samples': [(1, 0)]
        },
    },

    'integration_weights': {
        'cos': {
            'cosine': [(1, MEASUREMENT_PULSE_DURATION)],
            'sine': [(0, MEASUREMENT_PULSE_DURATION)],
        },
        'sin': {
            'cosine': [(0, MEASUREMENT_PULSE_DURATION)],
            'sine': [(1, MEASUREMENT_PULSE_DURATION)],
        },
        'x_const': {
            'cosine': [(CONST_INTEGRATION_WEIGHT, MEASUREMENT_PULSE_DURATION)],
            'sine': [(0, MEASUREMENT_PULSE_DURATION)],
        },
        'x_const_1us': {
            'cosine': [(1/1000, MEASUREMENT_PULSE_DURATION)],
            'sine': [(0, MEASUREMENT_PULSE_DURATION)],
        },
        'y_const': {
            'cosine': [(0, MEASUREMENT_PULSE_DURATION)],
            'sine': [(CONST_INTEGRATION_WEIGHT, MEASUREMENT_PULSE_DURATION)],
        },
        'x_const_1ms': {
            'cosine': [(CONST_INTEGRATION_WEIGHT, int(1e6))],
            'sine': [(0, int(1e6))],
        },
        'y_const_1ms': {
            'cosine': [(0, int(1e6))],
            'sine': [(CONST_INTEGRATION_WEIGHT, int(1e6))],
        },
    },
    'mixers': {
        'mixer_qe1': [
            {
                'intermediate_frequency': sideband_freqs['qe1'],
                'lo_frequency': LO_Q1,
                'correction': [1.0, 0.0, 0.0, 1.0]
            },
        ],
        'mixer_q1': [
            {
                'intermediate_frequency': larmor_freqs['Q1'],
                'lo_frequency': LO_Q1,
                'correction': [1.0, 0.0, 0.0, 1.0]
            },
        ],
        'mixer_y1': [
            {
                'intermediate_frequency': larmor_freqs['Q1'],
                'lo_frequency': LO_Q1,
                'correction': [0.0, 1.0, 1.0, 0.0]
            },
        ],
        'mixer_q2': [
            {
                'intermediate_frequency': larmor_freqs['Q2'],
                'lo_frequency': LO_Q2,
                'correction': [1.0, 0.0, 0.0, 1.0]
            },
        ],
        'mixer_y2': [
            {
                'intermediate_frequency': larmor_freqs['Q2'],
                'lo_frequency': LO_Q1,
                'correction': [0.0, 1.0, 1.0, 0.0]
            },
        ],
        'mixer_q3': [
            {
                'intermediate_frequency': larmor_freqs['Q3'],
                'lo_frequency': LO_Q3,
                'correction': [1.0, 0.0, 0.0, 1.0]
            },
        ],
        'mixer_q4': [
            {
                'intermediate_frequency': larmor_freqs['Q4'],
                'lo_frequency': LO_Q4,
                'correction': [1.0, 0.0, 0.0, 1.0]
            },
        ],
        'mixer_q5': [
            {
                'intermediate_frequency': larmor_freqs['Q5'],
                'lo_frequency': LO_Q5,
                'correction': [1.0, 0.0, 0.0, 1.0]
            },
        ],
        'mixer_q6': [
            {
                'intermediate_frequency': larmor_freqs['Q6'],
                'lo_frequency': LO_Q6,
                'correction': [1.0, 0.0, 0.0, 1.0]
            },
        ],
        'mixer_q7': [
            {
                'intermediate_frequency': larmor_freqs['Q7'],
                'lo_frequency': LO_Q7,
                'correction': [1.0, 0.0, 0.0, 1.0]
            },
        ],
        'mixer_q8': [
            {
                'intermediate_frequency': larmor_freqs['Q8'],
                'lo_frequency': LO_Q8,
                'correction': [1.0, 0.0, 0.0, 1.0]
            },
        ],
    }
}



"""Module containing example measurement configuration for eight dot device"""
from qcodes import validators

from arbok_driver.parameter_types import (
    Time, Voltage, Frequency, List, Int
)
from arbok_driver.examples.configurations import (
    square_pulse_conf,
    square_pulse_scalable_conf
)
device_config = {
    'parameters': {
        'iteration': {
            'type': Int,
            'value': int(1),
            'vals': validators.Ints(),
        },
        'gate_elements': {
            'type': List,
            'value': [
                'P1', 'J1',
                'P2',
            ]
        },
        'readout_elements': {
            'type': List,
            'value': [
                'SET1',
                'SET2',
            ]
        },
        'qubit_elements': {
            'type': List,
            'value': [
                'Q1', 'Q2',
                'Q3', 'Q4',
                'Q5', 'Q6',
                'Q7', 'Q8',
                ]
        },
        'f_larmor': {
            'type': Frequency,
            'label': 'resonance frequency',
            'elements': {
                'Q1': int(27.85e6),
                'Q2': int(30.1e6),
                'Q3': int(26.7e6),
                'Q4': int(48.1e6),
                'Q5': int(45.342e6),
                'Q6': int(50.340e6),
                'Q7': int(45.4e6),
                'Q8': int(42.8e6),
                'qe1': int(1e6),
            },
        },
        't_2pi': {
            'type': Time,
            'label': '2pi time',
            'elements': {
                'Q1': int(2500),
                'Q2': int(2283),
                'Q3': int(1000),
                'Q4': int(1600),
                'Q5': int(1642),
                'Q6': int(1604),
                'Q7': int(1136),
                'Q8': int(1923),
            },
        },
        'v_home': {
            'type': Voltage,
            'label': 'home voltage point',
            'elements': {
                'P1': 0, 'J1': 0,
                'P2': 0, 'J2': 0,
                'P3': 0, 'J3': 0,
                'P4': 0, 'J4': 0,
                'P5': 0, 'J5': 0,
                'P6': 0, 'J6': 0,
                'P7': 0, 'J7': 0,
                'P8': 0,
            }
        },
        'v_control': {
            'type': Voltage,
            'label': 'control voltage point',
            'elements': {
                'P1': 0.0, 'J1': -0.05,
                'P2': 0.0, 'J2': -0.,
                'P3': 0.2, 'J3': -0.2,
                'P4': 0.2, 'J4': 0,
                'P5': -0.017, 'J5': -0.19,
                'P6': 0.017, 'J6': 0,
                'P7': 0, 'J7': -0.63,
                'P8': 0, #'LCB': 0
            }
        },
        'v_exchange': {
            'type': Voltage,
            'label': 'Exchange voltage point',
            'elements': {
                'P1': 0, 'J1': 0.01,
                'P2': 0, 'J2': 0.02,
                'P3': 0, 'J3': 0.03,
                'P4': 0, 'J4': 0.04,
                'P5': 0, 'J5': 0.05,
                'P6': 0, 'J6': 0.06,
                'P7': 0, 'J7': 0.07,
                'P8': 0,
            }
        },
    },
    'default_sequence_configs': {
        'square_pulse': {
            'config': square_pulse_conf
        },
        'square_pulse_custom': {
            'config': square_pulse_scalable_conf
        },
    },
}



"""
Configuration for a eight quantum dot device initializing the qubits parity state
"""
from arbok_driver.parameter_types import (
    Time, Voltage, List
)

from arbok_driver.examples.sub_sequences.parity_initialization import ParityInit

parity_init_conf = {
    'sequence': ParityInit,
    'parameters': {
        'gate_elements': {
            'type': List,
            'value': [
                'P1', 'J1',
                'P2',
                'P7', 'J7',
                'P8',
                ]
            },
        't_ramp_to_t1': {
            'type': Time,
            'value': int(1e3/4)
            },
        't_wait_at_t1': {
            'type': Time,
            'value': int(200e3/4)
            },
        't_ramp_to_crossing': {
            'type': Time,
            'value': int(1e3/4)
            },
        't_wait_before_crossing': {
            'type': Time, 
            'value': int(9571)
            },
        't_ramp_over_crossing': {
            'type': Time,
            'value': int(141)
            },
        't_wait_after_crossing': {
            'type': Time,
            'value': int(1e3/4)
            },
        'tHomeRamp': {
            'type': Time,
            'value': int(600)
            },
        'v_home': {
            'type': Voltage,
            'elements': {
                    'P1': 0, 'J1': 0,
                    'P2': 0,
                    'P7': 0, 'J7': 0,
                    'P8': 0,
            }
        },
        'v_t1_point': {
            'type': Voltage,
            'elements': {
                    'P1': 0.06, 'J1': -0.085,
                    'P2': -0.06,
                    'P7': -0.07, 'J7': -0.05,
                    'P8': 0.07,
            }
        },
        'v_pre_crossing': {
            'type': Voltage,
            'elements': {
                    'P1': 0.0435, 'J1': -0.05,
                    'P2': -0.0435,
                    'P7': 0.029, 'J7': 0.067,
                    'P8': -0.029,
            }
        },
        'v_delta': {
            'type': Voltage,
            'elements': {
                    'P1': -0.03, 'J1': 0,
                    'P2': 0.03,
                    'P7': -0.02, 'J7': 0,
                    'P8': 0.02,
            }
        },
    }
}


"""Module containing configurations for SquarePulse(Scalable) classes"""
from arbok_driver.parameter_types import (
    Amplitude, List, String, Time, Voltage
)

from arbok_driver.examples.sub_sequences import SquarePulse, SquarePulseScalable

square_pulse_conf = {
    "sequence": SquarePulse,
    "parameters": {
        'amplitude': {'type': Amplitude, 'value': 1.5},
        'element': {'type': String, 'value': 'P1'},
        't_ramp': {'type': Time, 'value': int(100)},
        't_square_pulse': {'type': Time, 'value': int(1000)}
    }
}

square_pulse_scalable_conf = {
    "sequence": SquarePulseScalable,
    "parameters": {
        'amplitude': {'type': Amplitude, 'value': 1.5},
        't_ramp': {'type': Time, 'value': int(100)},
        'sticky_elements': {'type': List, 'value': ['P1', 'P2', 'P3']},
        't_square_pulse': {'type': Time, 'value': int(1000)},
        'v_home': {
            'type': Voltage,
            'elements': {
                'P1': 0.0,
                'P2': 0.0,
                'P3': 0.0
            }
        },
        'v_square': {
            'type': Voltage,
            'elements': {
                'P1': 0.1,
                'P2': -0.2,
                'P3': 0.2
            }
        }
    }
}


"""Module containing abstract class for experiments"""
from abc import ABC, abstractmethod
from typing import Mapping
from arbok_driver.device import Device

class Experiment(ABC):
    """
    Abstract class describing a type of experiment to run on the QM.
    This enforces a certain order of sequences that only requires certain
    input parameters or configurations to be set.
    E.g. for a rabi experiment, the order of sequences is:
        1. Initialization of quantum state
        2. Rabi pulse
        3. Readout of quantum state
    """
    _name: str
    def __init__(self, **configs_to_prepare: str | Mapping[str, dict] | None):
        self.configs: dict = {}
        self.configs_to_prepare: Mapping = configs_to_prepare

    @property
    @abstractmethod
    def sequences_config(self) -> dict:
        """Sequences to be run within program uploaded to the OPX"""
        pass

    @property
    def name(self) -> str:
        """Name of the experiment to run"""
        return self._name

    def __call__(self) -> dict:
        """Returns the sequences to be run"""
        return self.sequences_config

    def get_sequences_config(self, device: Device) -> dict:
        self.configs = self.get_device_specific_subsequences_dict(device)
        return self.sequences_config

    def get_device_specific_subsequences_dict(self, device: Device):
        """
        Prepare configurations for the experiment by updating the default configs.

        Args:
            device (dict): Device to use for default sequences
        """
        configs = device.default_sequence_configs
        if not isinstance(self.configs_to_prepare, dict):
            raise TypeError(
                f"{self.name}'s 'configs_to_prepare' must be of type dict"
                f" is type: {type(self.configs_to_prepare)}."
            )
        for name, config in self.configs_to_prepare.items():
            if isinstance(config, dict):
                configs[name] = config
            elif isinstance(config, str):
                default_name = config
                if default_name in configs:
                    configs[name] = configs[config]
                else:
                    raise KeyError(
                        f"Default config '{config}' not found in default configs."
                        f" Of device: {device.name}"
                        )
            elif config is None:
                if name not in configs:
                   configs[name] = {}
        return configs



"""Initialize measurement runners"""
from .measurement_runner_base import MeasurementRunnerBase
from .qcodes_measurement_runner import QCodesMeasurementRunner
from .native_measurement_runner import NativeMeasurementRunner



import importlib
import sys
import pkgutil
from types import ModuleType
from typing import Any, Dict, List


class _NamespaceNode:
    """Container for dynamically attached attributes."""

    def __init__(self) -> None:
        self.__dict__: Dict[str, Any] = {}

    def __repr__(self) -> str:
        return f"<NamespaceNode {list(self.__dict__.keys())}>"


class Ekans:
    """
    Ekans object providing a dynamic, reloadable view of a package.

    Attributes from submodules are flattened one level up, enabling:

        ekans.analysis.DecayFit

    instead of:

        ekans.analysis.decay_fit.DecayFit

    The module hierarchy is preserved internally using namespace nodes.
    """

    _root_module: ModuleType

    def __init__(self, root_module: ModuleType) -> None:
        if not isinstance(root_module, ModuleType):
            raise TypeError(
                f"root_module must be a Python module, got {type(root_module)}"
            )
        self._root_module = root_module
        self.reload_modules()

    def reload_modules(self) -> None:
        """Reload all modules and rebuild the attribute tree."""
        module_names: List[str] = self._collect_module_names()
        module_names.sort(key=lambda name: name.count("."), reverse=True)

        for name in module_names:
            try:
                if name in sys.modules:
                    module = sys.modules[name]
                    if getattr(module, "__spec__", None) is not None:
                        importlib.reload(module)
                else:
                    importlib.import_module(name)  # ← THIS is what you're missing

            except Exception as e:
                print(f"Error importing {name}: {e}")
                raise

        self._build_attribute_tree(module_names)

    def _collect_module_names(self) -> List[str]:
        root = self._root_module
        module_names: List[str] = [root.__name__]

        if hasattr(root, "__path__"):
            for module_info in pkgutil.walk_packages(
                root.__path__, prefix=root.__name__ + "."
            ):
                module_names.append(module_info.name)

        return module_names

    def _build_attribute_tree(self, module_names: List[str]) -> None:
        for key in list(self.__dict__.keys()):
            if not key.startswith("_"):
                delattr(self, key)

        root_name = self._root_module.__name__
        root_prefix = root_name + "."

        for full_name in module_names:
            module = sys.modules.get(full_name)
            if not isinstance(module, ModuleType):
                continue

            # Strip root prefix
            if full_name == root_name:
                parts: List[str] = []
            else:
                parts = full_name[len(root_prefix):].split(".")

            parent: Any = self

            for part in parts[:-1]:
                if not hasattr(parent, part):
                    setattr(parent, part, _NamespaceNode())
                parent = getattr(parent, part)

            if not parts:
                node = self
            else:
                name = parts[-1]

                if hasattr(module, "__path__"):
                    if not hasattr(parent, name):
                        setattr(parent, name, _NamespaceNode())
                    node = getattr(parent, name)
                else:
                    node = parent

            self._attach_module_attributes(node, module)

    def _attach_module_attributes(
        self, node: _NamespaceNode, module: ModuleType
    ) -> None:
        public_names = getattr(module, "__all__", None)

        if public_names is None:
            public_names = [
                name for name in dir(module)
                if not name.startswith("_")
            ]

        # Attach module's own attributes
        for attr_name in public_names:
            try:
                value = getattr(module, attr_name)
            except AttributeError:
                continue
            setattr(node, attr_name, value)

        # also pull from direct submodules if this is a package
        if hasattr(module, "__path__"):
            for submodule in self._get_submodules(module):
                sub_public = getattr(submodule, "__all__", None)

                if sub_public is None:
                    sub_public = [
                        name for name in dir(submodule)
                        if not name.startswith("_")
                    ]

                for attr_name in sub_public:
                    try:
                        value = getattr(submodule, attr_name)
                    except AttributeError:
                        continue

                    # avoid overwriting existing names
                    if not hasattr(node, attr_name):
                        setattr(node, attr_name, value)

    def _get_submodules(self, module: ModuleType) -> List[ModuleType]:
        prefix = module.__name__ + "."
        return [
            m for name, m in sys.modules.items()
            if name.startswith(prefix)
            and isinstance(m, ModuleType)
            and name.count(".") == prefix.count(".")  # direct children only
        ]


import copy
from dataclasses import dataclass
import re

import numpy as np
import pytest
from qm import generate_qua_script

from arbok_driver import Measurement, SubSequence, ParameterClass
from arbok_driver.examples.configurations import (
    divider_config
)
from arbok_driver.examples.sub_sequences import (
    SquarePulseScalable, ParityInit, ParityRead
    )
from arbok_driver.examples.configurations import (
    parity_init_conf, parity_read_conf
)
from arbok_driver.parameter_types import (
    Amplitude, List, Time, Voltage,
)

@dataclass(frozen=True)
class SubSequenceParameterClass(ParameterClass):
    par1: Amplitude

class UserSubSequence(SubSequence):
    PARAMETER_CLASS = SubSequenceParameterClass

def test_measurement_init(arbok_driver) -> None:
    """Tests if sequence is correctly initialized"""
    seq = Measurement(arbok_driver, 'mock_measurement')
    assert seq.name == f'{arbok_driver.name}_mock_measurement'

def test_measurement_sub_sequence_addition(mock_measurement) -> None:
    """Tests if subsequence is correctly added to sequence"""
    config_1 = {
        'parameters':{
            'par1': {'type': Time, 'value': int(100)},
            'par2': {'type': Time, 'value': int(10)},
            'par3': {'type': Voltage, 'value': 1.1},
            'v_home': {'type': Amplitude, 'elements': {
                'P1': 0.5, 'J1': 0.6, }},
        }
    }
    # SubSequence.PARAMETER_CLASS = SubSequenceParameterClass
    seq1 = UserSubSequence(
        parent = mock_measurement,
        name = 'seq1',
        sequence_config = config_1
        )
    assert len(mock_measurement.submodules) == 1
    assert len(mock_measurement.sub_sequences) == 1
    assert mock_measurement.sub_sequences[0] == seq1
    assert hasattr(mock_measurement, 'seq1')
    assert mock_measurement.seq1 == seq1

def test_set_sweeps(empty_sub_seq_1, mock_measurement) -> None:
    """Tests if sweeps are correctly set"""
    mock_measurement.set_sweeps(
        {empty_sub_seq_1.par1: np.arange(10)},
        {empty_sub_seq_1.par2: np.arange(5)}
        )
    assert len(mock_measurement.sweeps) == 2
    assert mock_measurement.sweeps[0].length == 10
    assert mock_measurement.sweeps[1].length == 5
    assert len(mock_measurement.sweeps[0].parameters) == 1
    assert mock_measurement.sweep_size == 50

def test_set_uneven_sweeps(empty_sub_seq_1, mock_measurement) -> None:
    with pytest.raises(ValueError):
        mock_measurement.set_sweeps(
            {
                empty_sub_seq_1.par1: np.arange(10),
                empty_sub_seq_1.par2: np.arange(5)
            }
            )
        
def test_non_dict_sweeps(empty_sub_seq_1, mock_measurement) -> None:
    with pytest.raises(TypeError):
        mock_measurement.set_sweeps(6, 'test')

def test_non_sequence_parameter_sweeps(empty_sub_seq_1, mock_measurement) -> None:
    with pytest.raises(TypeError):
        mock_measurement.set_sweeps({'test': np.arange(10)})

def test_zero_length_sweeps(empty_sub_seq_1, mock_measurement) -> None:
    with pytest.raises(ValueError):
        mock_measurement.set_sweeps({empty_sub_seq_1.par1: np.arange(0)})

def test_warning_for_duplicate_value_sweeps(
        empty_sub_seq_1, mock_measurement) -> None:
    with pytest.warns(UserWarning):
        mock_measurement.set_sweeps({empty_sub_seq_1.par1: [1, 1, 2, 3, 4]})

def test_qua_program_compilation_w_linear_sweeps(
        empty_sub_seq_1, mock_measurement) -> None:
    """Tests whether the qua code is compiled correctly"""
    mock_measurement.set_sweeps(
        {empty_sub_seq_1.par1: np.arange(0, 10, 1)},
        {empty_sub_seq_1.v_home_P1: np.arange(-0.1, 0.1, 0.01)}
        )
    qua_prog_str = mock_measurement.get_qua_program_as_str()
    qua_prog = mock_measurement.get_qua_program()
    qua_prog_str = generate_qua_script(qua_prog)
    # we expect 5 declares: 2x2 (2 per parameter -> iterator + qua_var)
    # + 1 to track the shots (always there!)
    assert len(re.findall(r'declare\(', qua_prog_str)) == 5
    # we expect 8:
    # 4 =2x2 (2 per parameter, initial val and increment)
    # +2 for shot tracker (init and increment)
    # +2 for each sweep axis
    assert len(re.findall('assign', qua_prog_str)) == 8
    # we expect 2 one per sweep axis
    assert len(re.findall('while_', qua_prog_str))== 2

def test_qua_program_compilation_w_non_linear_sweeps(
        empty_sub_seq_1, mock_measurement) -> None:
    """Tests whether the qua code is compiled correctly"""
    mock_measurement.set_sweeps(
        {empty_sub_seq_1.par2: np.arange(0, 10, 1)},
        {empty_sub_seq_1.v_home_P1: np.random.rand(8)}
        )
    qua_prog_str = mock_measurement.get_qua_program_as_str()
    qua_prog = mock_measurement.get_qua_program()
    qua_prog_str = generate_qua_script(qua_prog)
    # we expect 5 declares: 2x2 (2 per parameter -> iterator + qua_var)
    # +1 extra for declaring the explicit array (of non-lin sweep)
    # + 1 to track the shots (always there!)
    assert len(re.findall(r'declare\(', qua_prog_str)) == 6
    # we expect 8:
    # 4 =2x2 (2 per parameter, initial val and increment)
    # +2 for shot tracker (init and increment)
    # +2 for each sweep axis
    assert len(re.findall('assign', qua_prog_str)) == 8
    # we expect 2 one per sweep axis
    assert len(re.findall('while_', qua_prog_str)) == 2

def test_qua_program_compilation_simple_square_pulse(
        square_pulse, mock_measurement) -> None:
    qua_prog_str = mock_measurement.get_qua_program_as_str()
    qua_prog = mock_measurement.get_qua_program()
    qua_prog_str = generate_qua_script(qua_prog)
    print(qua_prog_str)
    assert len(re.findall(r'"ramp"\*amp', qua_prog_str)) == 2

@pytest.mark.parametrize("nr_gates", [0, 1, 2, 4])
def test_qua_program_compilation_scalable_square_pulse(
    mock_measurement,
    nr_gates,
    ) -> None:
    conf = {
        "parameters": {
            't_ramp': {'type': Time, 'value': int(100)},
            'sticky_elements': {'type': List, 'value': []},
            't_square_pulse': {'type': Time, 'value': int(1000)},
            'v_home': {'type': Voltage, 'elements': {}},
            'v_square': {'type': Voltage, 'elements': {}}
        }
    }
    init_conf = copy.deepcopy(conf)
    init_conf["parameters"]["v_home"]["elements"] = {
        f"P{i}": 0 for i in range(nr_gates)
    }
    init_conf["parameters"]["v_square"]["elements"] = {
        f"P{i}": 0.1 * (1 + i) for i in range(nr_gates)
    }
    init_conf["parameters"]["sticky_elements"]["value"] = [
        f"P{i}" for i in range(nr_gates)
    ]
    square_pulse = SquarePulseScalable(
        mock_measurement,
        f"square_pulse_{nr_gates}",
        init_conf,
    )
    for i in range(nr_gates):
        gate = f"P{i}"
        if gate in divider_config:
            scale = divider_config[gate]['division']
        else:
            scale = 1
        assert square_pulse.arbok_params.v_home[gate].qua == 0
        assert square_pulse.arbok_params.v_square[gate].qua == 0.1 * (1 + i)*scale
    qua_prog_str = mock_measurement.get_qua_program_as_str()
    nr_of_plays = len(
        list(re.finditer(r'"unit_ramp"\*amp', qua_prog_str))
    )
    assert nr_of_plays == 2 * nr_gates

    
def test_add_subsequences_from_dict(mock_measurement) -> None:
    mock_measurement.add_subsequences_from_dict(
        {
            'paraity_init_even': {'config': parity_init_conf},
            'parity_read_even': {'config': parity_read_conf},
            'parity_init_odd': {
                'config': parity_init_conf
            },
            'parity_read_odd': {
                'sequence': ParityRead,
                'config': parity_read_conf
            }
        }
    )
    assert len(mock_measurement.sub_sequences) == 4
    qua_prog_str = mock_measurement.get_qua_program_as_str()



"""Module containing the MeasurementRunner class."""
from __future__ import annotations
from typing import TYPE_CHECKING
import logging
import time
import uuid
import io
import json
import copy

import numpy as np
import xarray as xr
from sqlalchemy.orm import Session
from qm import generate_qua_script

from arbok_driver.sweep import Sweep
from arbok_driver.sqlalchemy_classes import SqlRun
from .measurement_runner_base import MeasurementRunnerBase
if TYPE_CHECKING:
    from arbok_driver.measurement import Measurement
    from qcodes.parameters.parameter import ParameterBase

TIME_DIMENSION = "batch_timestamp"

class NativeMeasurementRunner(MeasurementRunnerBase):
    """
    Helper class constructing QCoDeS measurement loops
    """

    def __init__(
        self,
        measurement: Measurement,
        ext_sweep_list: list[dict] | None = None,
        register_all: bool = True
        ):
        super().__init__(
            measurement=measurement,
            ext_sweep_list=ext_sweep_list,
            register_all=register_all
        )
        self.sql_run: SqlRun | None = None
        self.minio_data_store = None
        self.bucket_entry_name = None
        self.arbok_driver.check_db_engine_and_bucket_connected()

    def _prepare_measurement(self) -> None:
        """
        Prepare the measurement for running. 
            1) creating entry in database
            2) saving qua program as metadata
            3) saving qcodes snapshot as metadata
        """
        self._add_run_row_to_database()
        self._save_qua_program_as_metadata()
        self._save_qcodes_snapshot_as_metadata()
        self._initialize_data_store()

    def _handle_keyboard_interrupt(self) -> None:
        """
        Handle keyboard interrupt during measurement.
        """
        self.measurement._set_dataset(
            self.measurement.driver.get_data_from_id(self.run_id))

    def _wrap_up_measurement(self) -> None:
        """
        Wrap up the measurement after running.
        """
        self.measurement._set_dataset(
            self.measurement.driver.get_data_from_id(self.run_id))

    def _save_results(self):
        """
        Saves the results of the measurement to the datasaver.
        
        Args:
            datasaver (DataSaver): The QCoDeS DataSaver object to save results to.
        """
        ### Coordinate register to correcltly index incoming opx batches
        ### OPX coords are always the same for each batch
        ### External coords change with each batch
        batch_coords = {}
        for param, value in self.external_param_values.items():
            dim = self.ext_coords[param.register_name][0]
            batch_coords[param.register_name] = (dim, np.array([value]))
        batch_coords.update(self.opx_coords)

        dataset = xr.Dataset(coords = batch_coords)
        for _, gettable in self.measurement.gettables.items():
            result_np = gettable.fetch_results()
            result_xr = xr.DataArray(
                data = self._bring_result_to_shape(result_np),
                dims = self.dims, 
                coords = batch_coords
                )
            dataset[gettable.register_name] = result_xr
        self._save_dataset_to_store(dataset)
        self._update_run_row_in_database()
        logging.debug("Results saved")

    def _add_run_row_to_database(self) -> None:
        """
        Adds a new run row to the database for the current measurement.
        """
        with Session(self.arbok_driver.database_engine) as session:
            self.sql_run = SqlRun(
                exp_id=1,
                uuid=str(uuid.uuid4()),
                name=self.measurement.qc_measurement_name,
                # device=self.measurement.device.name,
                coords = list(self.coords.keys()),
                sweeps = {}, #self.coords,
                setup='rf2v',
                start_time=time.time(),
            )
            self.measurement._set_run_id(self.sql_run.run_id)
            session.add(self.sql_run)
            session.commit()
            self.run_id = int(self.sql_run.run_id)
            print("Inserted run with ID:", self.sql_run.run_id)
            self.bucket_entry_name = f"{self.sql_run.run_id}_{self.sql_run.uuid}"
            self.minio_data_store = self.arbok_driver.minio_filesystem.get_mapper(
                f"dev/{self.bucket_entry_name}/data.zarr")
            self.minio_metadata_store = self.arbok_driver.minio_filesystem.get_mapper(
                f"dev/{self.bucket_entry_name}/metadata")

    def _update_run_row_in_database(self) -> None:
        """
        Updates the run row in the database after each batch.
        """
        with Session(self.arbok_driver.database_engine) as session:
            sql_run = session.get(SqlRun, self.run_id)
            if sql_run is None:
                    raise ValueError(
                        f"SqlRun with id={self.run_id} not found at batch {self.batch_count}")
            sql_run.batch_count = self.batch_count + 1
            sql_run.result_count += self.measurement.sweep_size
            session.commit()

    def _save_dataset_to_store(self, dataset: xr.Dataset) -> None:
        """
        Saves the given dataset to the MinIO store.

        Args:
            dataset (xr.Dataset): The dataset to be saved.
        """
        dataset.to_zarr(
            self.minio_data_store,
            mode="r+",
            region = 'auto'
            )

    def _save_qua_program_as_metadata(self) -> None:
        """
        Saves the QUA program as metadata to the binary bucket.
        """
        program_str = generate_qua_script(
            self.measurement.qua_program,
            self.measurement.device.config
        )
        program_str_binary = io.BytesIO(program_str.encode("utf-8"))
        self.minio_metadata_store["qua_program.py"] = program_str_binary.getvalue()

    def _save_qcodes_snapshot_as_metadata(self) -> None:
        """
        Saves the QCoDeS snapshot as metadata to the binary bucket.
        """
        driver = self.measurement.driver
        if not hasattr(driver, 'station'):
            print("Not saving QCoDeS snapshot, no station found.")
            return
        if driver.station is None:
            print("Not saving QCoDeS snapshot, no station found.")
            return
        snapshot_dict = driver.station.snapshot(update=False)
        snapshot_bytes = json.dumps(snapshot_dict, indent=2).encode("utf-8")
        self.minio_metadata_store["qcodes_snapshot.json"] = snapshot_bytes

    def _initialize_data_store(self) -> None:
        """
        Initializes the data store by creating an empty xarray dataset with the
        correct coordinates and dimensions.
        """  
        result_vars = {}
        for gettable in self.measurement.gettables.values():
            result_shape = tuple(
                (len(self.coords[dim][1]) for dim in self.dims))
            result_vars[gettable.register_name] = xr.DataArray(
                data = np.full(shape = result_shape, fill_value=np.nan),
                dims = self.dims,
                coords = self.coords
            )
        empty_dataset = xr.Dataset(
            data_vars = result_vars, coords = self.coords)
        for coord_name in empty_dataset.coords:
            empty_dataset.coords[coord_name].attrs['units'] = self.units[coord_name]
        empty_dataset.to_zarr(self.minio_data_store, mode='w')



""" Module containing GettableParameter class """
from __future__ import annotations
from typing import Type, TYPE_CHECKING

from qm import qua

from .gettable_parameter_base import GettableParameterBase

if TYPE_CHECKING:
    from arbok_driver.read_sequence import ReadSequence
    from qm.qua._expressions import QuaVariable

class GettableParameter(GettableParameterBase):
    """
    Gettableparameter class handling sclar results from an abstract readout
    """
    qua_result_var: QuaVariable

    def __init__(
            self,
            name: str,
            read_sequence: ReadSequence,
            var_type: Type[int | bool | qua.fixed],
            **kwargs
            ) -> None:
        """
        Constructor class for ReadSequence class
        Args:
            name (dict): name of the GettableParameter
            read_sequence (ReadSequence): Readout class summarizing data streams
                and variables
            var_type (int | bool | qua.fixed): Type of qua variable to use
            **kwargs
        """
        super().__init__(
            name = name,
            read_sequence = read_sequence,
            var_type = var_type,
            **kwargs
            )
        self.reset_measuerement_attributes()

    def qua_declare_variables(self):
        """Saves acquired results to qua stream"""
        self.qua_result_var = qua.declare(self.var_type)
        self.qua_stream = qua.declare_stream()

    def qua_save_variables(self):
        qua.save(self.qua_result_var, self.qua_stream)

    def set_raw(self, *args, **kwargs) -> None:
        """Empty abstract `set_raw` method. Parameter not meant to be set"""
        raise NotImplementedError("GettableParameters are not meant to be set")

    def reset_measuerement_attributes(self):
        """Resets all job specific attributes"""
        super().reset_measuerement_attributes()
        self.qua_result_var = None



"""Module containing ChoppedReadout class"""
from dataclasses import dataclass

from qm import qua
from qm.qua._expressions import QuaVariable, QuaBinaryOperation
from arbok_driver import (
    ReadSequence, AbstractReadout, arbok, Signal, ParameterClass
)
from arbok_driver.parameter_types import (
    Voltage, Time, Int, ParameterMap
)

@dataclass(frozen = True)
class DcChoppedReadoutParameters(ParameterClass):
    v_home: ParameterMap[str, Voltage]
    v_chop: ParameterMap[str, Voltage]
    n_chops: Int
    chop_wait: Time

class DcChoppedReadout(AbstractReadout):
    """Abstract readout class for chopped readout"""
    PARAMETER_CLASS = DcChoppedReadoutParameters
    arbok_params: DcChoppedReadoutParameters

    n_chops: QuaVariable
    qua_chop_nr: QuaVariable
    n_chops_division: float | QuaBinaryOperation

    def __init__(
        self,
        name: str,
        read_sequence: ReadSequence,
        signal: Signal,
        save_results: bool,
        parameters: dict,
        readout_qua_elements: dict[str, str],
        ):
        """
        Constructor method for `ChoppedReadout` abstractreadout helper class
        
        Args:
            name (str): Name of the abstract readout. Will be added to
                respective signals like this
            read_sequence (arbok_driver.ReadSequence): ReadSequence in which this
                abstract readout is performed
            save_results (bool): Whether to save the results
            parameters (dict): Parameters to be added to the read sequence
            readout_qua_elements (dict[str, str]): Dict with signals as keys and
                qua readout elements as values
        """

        super().__init__(
            name = name,
            read_sequence = read_sequence,
            signal = signal,
            save_results = save_results,
            parameters = parameters
        )
        self.readout_qua_elements = readout_qua_elements

        self.gate_elements = list(self.arbok_params.v_chop.keys())
        self.elements = self.gate_elements + list(
            self.readout_qua_elements.values())

        self.ref_temp_vars: dict = {}
        self.read_temp_vars: dict = {}
        self.read_gettables: dict = {}
        self.ref_gettables: dict = {}
        self.diff_gettables: dict = {}
        self._create_gettables()

    def qua_declare_variables(self) -> None:
        """Declares all necessary qua variables for readout"""
        self.n_chops_division = 1/self.arbok_params.n_chops.qua
        super().qua_declare_variables()
        self.qua_chop_nr = qua.declare(int)
        for sub_readout, _ in self.readout_qua_elements.items():
            self.ref_temp_vars[sub_readout] = qua.declare(qua.fixed)
            self.read_temp_vars[sub_readout] = qua.declare(qua.fixed)

    def qua_measure(self) -> None:
        """Measures the given observables and assigns the result to the vars"""
        for sub_readout, _ in self.readout_qua_elements.items():
            qua.assign(self.diff_gettables[sub_readout].qua_result_var, 0.)
            qua.assign(self.ref_temp_vars[sub_readout], 0.)
            qua.assign(self.read_temp_vars[sub_readout], 0.)

        with qua.for_(
            var = self.qua_chop_nr,
            init = 0,
            cond = self.qua_chop_nr < self.arbok_params.n_chops.qua,
            update = self.qua_chop_nr + 1
            ):
            qua.align()
            arbok.ramp(
                elements= self.gate_elements,
                target = self.arbok_params.v_chop,
                operation = 'unit_ramp',
                )

            qua.align(*self.elements)
            qua.wait(self.arbok_params.chop_wait.qua, *self.elements)
            qua.align(*self.elements)

            for sub_readout, qua_element in self.readout_qua_elements.items():
                outputs = [
                    qua.integration.full(
                        'x_const',
                        self.ref_gettables[sub_readout].qua_result_var),
                    ]
                qua.measure('measure', qua_element, *outputs)
            qua.align(*self.elements)

            arbok.ramp(
                elements= self.gate_elements,
                reference = self.arbok_params.v_chop,
                target = self.arbok_params.v_home,
                operation = 'unit_ramp',
                )

            qua.align(*self.elements)
            qua.wait(self.arbok_params.chop_wait.qua, *self.elements)
            qua.align(*self.elements)

            for sub_readout, qua_element in self.readout_qua_elements.items():
                outputs = [
                    qua.integration.full('x_const', self.read_gettables[sub_readout].qua_result_var),
                    ]
                qua.measure('measure', qua_element, *outputs)
            qua.align(*self.elements)

            ### Calculate difference between read and reference point
            ### Add to total result
            for sub_readout, _ in self.readout_qua_elements.items():
                # Calculate the difference in qua
                qua.assign(
                    self.ref_temp_vars[sub_readout],
                    self.ref_temp_vars[sub_readout]
                    + self.ref_gettables[sub_readout].qua_result_var*self.n_chops_division
                )
                qua.assign(
                    self.read_temp_vars[sub_readout],
                    self.read_temp_vars[sub_readout]
                    + self.read_gettables[sub_readout].qua_result_var*self.n_chops_division
                )
        qua.align()
        ### Normalize the result and threshold it
        for sub_readout, _ in self.readout_qua_elements.items():
            ### Save accumulated ref and read to respective observables
            qua.assign(
                self.ref_gettables[sub_readout].qua_result_var,
                self.ref_temp_vars[sub_readout]
                )
            qua.assign(
                self.read_gettables[sub_readout].qua_result_var,
                self.read_temp_vars[sub_readout]
                )
            ### Calculate the difference and threshold it
            qua.assign(
                self.diff_gettables[sub_readout].qua_result_var,
                self.read_temp_vars[sub_readout] - self.ref_temp_vars[sub_readout]
                )
        qua.align(*self.elements)

    def _create_gettables(self) -> None:
        """
        Populates helper attributes for chopped readout class instance.

        Attributes:
            elements (list): List of elements to be chopped
            signals (list): List of signals on which results are saved
            ref__gettables (dict): Dict with signal names as keys and
                respective reference observables as values
            read__gettables (dict): Dict with signal names as keys and
                respective read observables as values
            chop_observables (dict): Dict with signal names as keys and
                respective chop observables as values
            state_observables (dict): Dict with signal names as keys and
                respective state observables as values
            observables (dict): Dict with FULL observable names as keys and
                ALL observables as values. Crucial for auto param declare
        """
        for sub_readout, _ in self.readout_qua_elements.items():
            read = self.read_gettables[sub_readout] = self.create_gettable(
                gettable_name = f"{sub_readout}_read",
                var_type = qua.fixed,
            )
            ref = self.ref_gettables[sub_readout] = self.create_gettable(
                gettable_name = f"{sub_readout}_ref",
                var_type = qua.fixed,
            )
            diff = self.diff_gettables[sub_readout] = self.create_gettable(
                gettable_name = f"{sub_readout}_diff",
                var_type = qua.fixed,
            )
            self.read_gettables[sub_readout] = read
            self.ref_gettables[sub_readout] = ref
            self.diff_gettables[sub_readout] = diff



"""Module containing GenericTuningInterface class."""
from __future__ import annotations
from collections.abc import Callable
import copy
import random
import time
from typing import Protocol, TYPE_CHECKING

from scipy.stats import qmc
import matplotlib.pyplot as plt
import xarray as xr
import numpy as np
from rich.progress import Progress
from rich import print
from IPython import display

if TYPE_CHECKING:
    from .arbok_driver import ArbokDriver
    from .measurement import Measurement
    from .parameters import SequenceParameter, GettableParameterBase

class CostFunction(Protocol):
    def __call__(
        self,
        results: dict[GettableParameterBase, np.ndarray]
    ) -> float:
        ...


class GenericTuningInterface:
    """
    Generic streaming interface for machine learning–based parameter tuning.

    This class coordinates measurement execution, parameter management,
    and cost evaluation via a pluggable ``CostStrategy``.

    Attributes:
        measurement (Measurement):
            Measurement object providing access to the experiment.
        driver (ArbokDriver):
            Driver associated with the measurement.
        parameter_dict (dict[str, dict]):
            Dictionary describing tunable parameters and their configurations.
        sweeps (list[dict[SequenceParameter, Sequence]]):
            List of sweep configurations defining remaining parameters are
            being varied
        gettables (dict[str, GettableParameterBase]): Gettables to be saved
        verbose (bool): Whether all optional printouts are shown
    """

    def __init__(
        self,
        measurement: Measurement,
        parameter_dicts: dict[str, dict],
        verbose: bool = False
    ) -> None:
        """
        Initialize the tuning interface.

        Args:
            measurement (Measurement):
                Measurement object that provides access to the experiment
                and its associated driver.
            parameter_dicts (dict[str, dict[SequenceParameter, float]]):
                Dictionary defining parameter groups and their bounds.
                Keys represent trivial parameter names or groups, and values
                map ``SequenceParameter`` objects to their bounds and factors.
            verbose (bool, optional):
                If ``True``, enables verbose logging during initialization and
                parameter setup. Defaults to ``False``.

        Raises:
            TypeError:
                If ``cost_strategy`` is not an instance of ``CostStrategy``.
        """
        self.measurement: Measurement = measurement
        self.driver: ArbokDriver = self.measurement.driver
        self._add_parameters(parameter_dicts, verbose=verbose)

        if not self.measurement.sweeps:
            raise ValueError(
                "No sweeps registered! Make sure to run measurement.set_sweeps(...)"
            )
        if not self.measurement.gettables:
            raise ValueError(
                "No gettables registered!"
                "Make sure to run measurement.register_gettables(...)"
            )
        self.gettables = self.measurement.gettables
        self.sweeps = self.measurement.sweeps

    def _add_parameters(self, parameter_dicts: dict, verbose: bool = False) -> None:
        """
        Adds parameters to be streamed to the QUA program.
        Parameters that depend on each other are added in the folowing way;
            'even_hotspot_detune': {
                interface.sequence.init_even.v_hotspot_P1: 1,
                interface.sequence.init_even.v_hotspot_P2: -1,
                },
        If added in this way, only the first parameter will be added to the 
        input stream, while all following once will use its qua variable as a
        reference with the respective factor.
        """
        self.parameter_dict = parameter_dicts
        self.input_stream_params = []
        self.bounds = {}
        for name, param_conf in parameter_dicts.items():
            for i, (parameter, factor) in enumerate(param_conf['qua_vars'].items()):
                if i == 0:
                    self.input_stream_params.append(parameter)
                    if verbose:
                        print(f"Adding input stream for {parameter.name} ({name})")
                else:
                    master_param = self.input_stream_params[-1]
                    def call(x = None, par = master_param, factor = factor):
                        if factor == 1:
                            return par(x)
                        else:
                            return par(x)*factor
                    parameter.call_method = call
                    if verbose:
                        print(
                            f"\tAdding {parameter.name} to {master_param.name}"
                            f"  input stream (factor: {factor}) ({name})")
            self.bounds[name] = param_conf['bounds']
        self.measurement.input_stream_parameters = self.input_stream_params

    def compile_connect_and_run(self, host_ip: str):
        """
        Compiles, connects and runs the parity readout sequences on device with
        given host ip.
        """
        if not self.driver.is_mock:
            self.driver.connect_opx(host_ip = host_ip)
            self.measurement.compile_qua_and_run()

    def run_parameter_set(
        self,
        input_param_dict: dict[SequenceParameter, float],
        progress_bar = None
        ) -> tuple[dict, dict]:
        """
        Runs the given parameter set an returns the current values for
        singlet and triplet init

        Args:
            input_param_dict (dict[SequenceParameter, float]): List of
                parameters to run
            progress_bar (Optional): Progress bar to update

        Returns:
            dict: All measured gettables for the parameter set
            dict: All parameters of the parameter set
        """
        if not isinstance(input_param_dict, dict):
            raise ValueError(
                "Input params must be a dict of SequenceParameters and float" 
                f" key value pairs. Are {type(input_param_dict)}")
        self.measurement.insert_single_value_input_streams(input_param_dict)
        if not self.driver.is_mock:
            self.driver.qm_job.resume()
        gettable_results = {}
        self.measurement.wait_until_result_buffer_full(
            progress_bar
            )
        results = self.measurement.fetch_all_results()
        for _, gettable in self.gettables.items():
            gettable_results[gettable] = results[gettable]
        saved_params = {}
        for param_name, value in zip(self.parameter_dict.keys(), input_param_dict.values()):
            saved_params[param_name] = value
        return gettable_results, saved_params

    def run_cross_entropy_sampler(
            self,
            populations: list,
            cost_strategy: CostFunction,
            select_frac: float = 0.3,
            plot_histograms: bool = False,
            sampling_params_to_plot: list | None = None
            ) -> xr.Dataset:
        """
        Runs the cross entropy method for the given populations.

        Args:
            populations (list[int]):
                List of population sizes for each iteration.

            cost_strategy (CostFunction):
                A callable that takes a dictionary mapping gettables to
                numpy arrays of measurement results and returns a scalar cost.

            select_frac (float):
                Fraction of best parameters to select for generating new bounds.

            plot_histograms (bool):
                If True, plot parameter distributions during sampling.

            sampling_params_to_plot (list[tuple[str, str]] | None):
                Parameter pairs to visualize during the sampling process.

        Returns:
            xr.Dataset: Dataset containing all gettables, parameters and
                rewards
        """
        all_rewards = []
        all_obs = {g: [] for g in self.gettables.values()}
        all_params = {param: [] for param in self.parameter_dict.keys()}
        all_bounds = {name: [] for name in self.bounds.keys()}
        current_bounds = copy.deepcopy(self.bounds)
        data_index = 0
        for population in populations:
            ### Sampling parameter sets and saving bounds
            print('Current bounds:\n', current_bounds)
            for param_name, bounds in current_bounds.items():
                all_bounds[param_name].append(bounds)
            sobol_samples = sobol_sampling(population, current_bounds)
            t0 = time.time()
            with Progress() as progress:
                task = progress.add_task(
                    "Sampling parameter sets", total=population)
                batch_task = progress.add_task(
                    "Sampling batch", total=self.measurement.sweep_size)
                total_nr = len(sobol_samples)
                ### Looping over all deviced parameter sets
                if plot_histograms:
                    fig, axs = plt.subplots(1, 2, figsize = (9,5))
                for i, x in enumerate(sobol_samples):
                    ### Running the parameter set
                    x = dict(zip(self.input_stream_params, x))
                    obs, par_dict = self.run_parameter_set(
                        x, progress_bar = (batch_task, progress))
                    r = cost_strategy(obs)
                    ### Saving the results
                    for param, value in par_dict.items():
                        all_params[param].append(value)
                    for gettable, value in obs.items():
                        all_obs[gettable].append(value)
                    all_rewards.append(r)
                    ### Updating the progress bar
                    progress.advance(task)
                    description = f"{i}/{total_nr} | "
                    description += f"Last SNR {np.max(r):.2f}, "
                    description += f"max: {np.max(all_rewards):.2f}"
                    progress.update(
                        task,
                        description = description
                        )
                    progress.refresh()
                    data_index += 1
                    if plot_histograms:
                        axs[0].cla()
                        axs[0].set_title('Best histogram ')
                        if len(all_rewards) == 0 or r > np.max(all_rewards):
                            for gettable, data in obs.items():
                                _ = axs[0].hist(data, label = gettable.name, alpha = 0.6)
                        axs[1].cla()
                        axs[1].set_title(f'Last histogram({i}/{total_nr})')
                        for gettable, data in obs.items():
                            _ = axs[1].hist(data, label = gettable.name, alpha = 0.6)

                        display.clear_output(wait=True)
                        display.display(plt.gcf())
                plt.close()
            print(f"Total time elapsed: {time.time()-t0:.0f}s")
            ### Updating the bounds for the next iteration
            dataset = self._merge_cem_data_into_xarray(
                all_rewards, all_obs, all_params)
            current_bounds = self._update_sobol_bounds(
                dataset,
                select_frac,
                population,
                sampling_params_to_plot,
                )
        ### Compressing data into xarray dataset and adding metadata
        dataset = dataset.assign_attrs(populations = populations)
        dataset = dataset.assign_attrs(bounds = bounds)
        return dataset

    def _merge_cem_data_into_xarray(
            self,
            all_rewards: list,
            all_obs: dict[GettableParameterBase, np.ndarray],
            all_params: dict[str, float]
            ):
        """
        Merges the data into an xarray dataset.
        
        Args:
            all_rewards (list): List of rewards for the current iteration
            all_obs (dict): Dict of gettable names and values
            all_params (dict): Dict of par names and values for the last population

        Returns:
            xr.Dataset: Dataset containing all data and metadata
        """
        nr_indices = len(all_rewards)
        dataset = xr.Dataset()
        
        ### Saving rewards
        dataset['rewards'] = xr.DataArray(
            np.array(all_rewards),
            coords = {'index':np.arange(nr_indices)},
            dims = ('index')
            )
        dataset['rewards'] = dataset.rewards.assign_attrs(type = 'reward')
        ### Saving gettables
        for gettable, data in all_obs.items():
            data = np.array(data)
            name = gettable.register_name
            dataset[name] = xr.DataArray(
                data,
                coords = {
                    'index':np.arange(nr_indices),
                    'shot_nr': np.arange(np.shape(data)[1])
                    },
                dims = ('index', 'shot_nr')
                )
            dataset[name] = dataset[name].assign_attrs(type = 'gettable')
        ### Saving parameters
        for par_name, data in all_params.items():
            data = np.array(data)
            dataset[par_name] = xr.DataArray(
                data,
                coords = {'index': np.arange(nr_indices)},
                dims = ('index')
                )
            dataset[par_name] = dataset[par_name].assign_attrs(type = 'parameter')
        ### Adding metadata
        dataset = dataset.assign_attrs(parameters = list(all_params.keys()))
        gettable_names = [g.register_name for g in all_obs.keys()]
        dataset = dataset.assign_attrs(gettables = gettable_names)
        return dataset

    def _update_sobol_bounds(
            self,
            dataset: xr.Dataset,
            select_frac: float,
            population_size: int,
            sampling_params_to_plot: list | None = None,):
        """
        Updates the bounds for the Sobol devicer.
        
        Args:
            rewards (list): List of rewards for the current iteration
            params (dict): Dict of par names and values for the last population
        """
        dataset = dataset.sel(index = dataset.index[-population_size:])
        nr_devices = int(np.ceil(select_frac*population_size))
        sorted_dataset = dataset.sortby(dataset.rewards)
        best_indices = sorted_dataset.index[-nr_devices:]

        new_bounds = {}
        for par_name in dataset.parameters:
            best_params = dataset[par_name].sel(index = best_indices)
            param_mean = float(best_params.mean().to_numpy())
            param_std = float(best_params.std().to_numpy())
            new_bounds[par_name] = (param_mean - param_std*1.5, param_mean + param_std*1.5)

        if sampling_params_to_plot is not None:
            nr_plots = len(sampling_params_to_plot)
            fig, axs = plt.subplots(1, nr_plots, figsize=(nr_plots*4, 5))
            if not isinstance(axs, list):
                axs = [axs]
            for i, (par1_name, par2_name) in enumerate(sampling_params_to_plot):
                param_bounds1 = new_bounds[par1_name]
                param_bounds2 = new_bounds[par2_name]
                axs[i].scatter(
                    dataset[par1_name].to_numpy(),
                    dataset[par2_name].to_numpy(),
                    color = 'blue', label = 'new bounds'
                    )
                axs[i].scatter(
                    dataset[par1_name].sel(index = best_indices).to_numpy(),
                    dataset[par2_name].sel(index = best_indices).to_numpy(),
                    color = 'red',
                    )
                axs[i].plot(
                    [
                        param_bounds1[0],
                        param_bounds1[1],
                        param_bounds1[1],
                        param_bounds1[0],
                        param_bounds1[0]
                    ],
                    [
                        param_bounds2[0],
                        param_bounds2[0],
                        param_bounds2[1],
                        param_bounds2[1],
                        param_bounds2[0]
                    ],
                    '-k')
                axs[i].set_xlabel(par1_name)
                axs[i].set_ylabel(par2_name)
            fig.tight_layout()
            plt.show()
        return new_bounds

    def _merge_data_into_xarray(self, index, all_rewards, all_obs, all_params):
        """Merges the data into an xarray dataset."""
        nr_indices = len(all_rewards)
        dataset = xr.Dataset()
        dataset['rewards'] = xr.DataArray(
            all_rewards,
            coords = {'index':np.arange(nr_indices)},
            dims = ('index')
            )
        dataset['params'] = xr.DataArray(
            np.array(all_params),
            coords = {'index':np.arange(nr_indices)},
            dims = ('index')
            )
        dataset['obs'] = xr.DataArray(
            np.array(all_obs),
            coords = {'index':np.arange(nr_indices),
                      'shot_nr': np.arange(self.measurement.sweep_size)},
            dims = ('index', 'shot_nr')
            )
        return dataset

def sobol_sampling(num_samples: int, bound_dict: dict) -> np.ndarray:
    """
    Generate Sobol sequence devices for the given parameters.
    
    Args:
        num_samples (int): Number of devices to generate.
        parameter_dict (dict): Dictionary containing the parameters to device.
            Must have bounds as a key for each parameter.
        
    Returns:
        dict: Dictionary containing the parameters as keys and the devices as values.
    """
    # Generate Sobol sequence devices and truncate devices to the desired number
    sobol_engine = qmc.Sobol(d=len(bound_dict), scramble=True)
    sobol_samples = sobol_engine.random_base2(m=int(np.ceil(np.log2(num_samples))))
    sobol_samples = np.array(random.sample(sobol_samples.tolist(), num_samples))

    # Scale devices to the desired domain
    for i, (_, config) in enumerate(bound_dict.items()):
        l_bound = config[0]
        u_bound = config[1]
        sobol_samples[:,i] = l_bound + (u_bound - l_bound) * sobol_samples[:,i]
    return sobol_samples



"""Module containing ParameterClass"""
from abc import ABC
from dataclasses import dataclass, fields

@dataclass(frozen = True)
class ParameterClass(ABC):
    """
    Parent for shared identity for the parameter classes every sub-sequence should own
    """
    def __len__(self) -> int:
        return len(fields(self))

@dataclass(frozen = True)
class EmptyParameterClass(ParameterClass):
    """Empty ParameterClass for all sub-sequences that require no parameters"""
    pass



"""Module containing Average child class inheriting from ReadoutPoint"""
from __future__ import annotations
from typing import TYPE_CHECKING

from qm import qua
from arbok_driver import (
    AbstractReadout,
    ReadSequence,
    Signal, 
)

if TYPE_CHECKING:
    from arbok_driver.parameters.gettable_parameter import GettableParameter

class DcAverage(AbstractReadout):
    """
    ReadoutPoint child class providing a `qua_measure` method to measure I and Q
    in a static averaging manner
    """
    def __init__(
        self,
        name: str,
        read_sequence: ReadSequence,
        signal: Signal,
        save_results: bool,
        parameters: dict,
        qua_element: str,
        ):
        """
        Constructor method of ReadoutPoint class
        
        Args:
            point_name (str): Name of readout point
            signal (Signal): Signal corresponding to readout point
            config (dict): List of readout points
        """
        super().__init__(
            name = name,
            read_sequence = read_sequence,
            signal = signal,
            save_results = save_results,
            parameters = parameters
        )
        self.signal = signal
        self.qua_element = qua_element
        self._create_gettables()

    def qua_measure(self):
        """
        Measures I and Q at the given read point in a static averaging manner
        """
        outputs = [
            qua.integration.full('x_const', self.current_gettable.qua_result_var),
            ]
        qua.measure('measure', self.qua_element, None, *outputs)

    def _create_gettables(self):
        self.current_gettable: GettableParameter = self.create_gettable(
            gettable_name = self.name,
            var_type = qua.fixed
        )



""" Module containing GettableParameter class """
from __future__ import annotations
from typing import TYPE_CHECKING, Sequence

from qm import qua
import numpy as np

from .gettable_parameter_base import GettableParameterBase

if TYPE_CHECKING:
    from arbok_driver.read_sequence import ReadSequence
    from qm.qua._expressions import QuaVariable, QuaArrayVariable
    from qcodes.parameters.parameter_base import ParameterBase
    from .sequence_parameter import SequenceParameter

class GettableParameterMulti(GettableParameterBase):
    """
    GettableParameterMulti class handling high dimensional results from an
    abstract readout
    """
    def __init__(
            self,
            name: str,
            read_sequence: ReadSequence,
            var_type: int | bool | qua.fixed,
            internal_setpoints: Sequence[SequenceParameter],
            **kwargs
            ) -> None:
        """
        Constructor class for ReadSequence class
        Args:
            name (dict): name of the GettableParameter
            read_sequence (ReadSequence): Readout class summarizing data streams
                and variables
            var_type (int | bool | qua.fixed): type of qua variable to be saved
            setpoints (Sequence[SequenceParameter]): sequence of setpoints
            **kwargs
        """
        super().__init__(
            name = name,
            read_sequence = read_sequence,
            var_type = var_type,
            **kwargs
            )
        self.reset_measuerement_attributes()
        self.internal_setpoints: tuple[SequenceParameter] = tuple(internal_setpoints)
        self.qua_var_index: QuaVariable | None = None
        self.length: int = self.get_length()

    def qua_declare_variables(self) -> None:
        """Declares the qua variables and streams for this gettable"""
        self.length = self.get_length()
        self.qua_var_index: QuaVariable = qua.declare(int)
        self.qua_result_array: QuaArrayVariable = qua.declare(
            self.var_type,
            size = self.length
        )
        self.qua_stream = qua.declare_stream()

    def qua_save_variables(self) -> None:
        """Saves acquired results to qua stream"""
        with qua.for_(
            var = self.qua_var_index,
            init = 0,
            cond = self.qua_var_index < self.length,
            update = self.qua_var_index + 1
            ):
            qua.save(
                self.qua_result_array[self.qua_var_index], self.qua_stream)

    def reset_measuerement_attributes(self):
        """Resets all job specific attributes"""
        super().reset_measuerement_attributes()
        self.qua_result_array = None

    def configure_from_measurement(self, setpoints: tuple[ParameterBase, ...]):
        """
        Configures the gettable parameter from the measurement object.
        This method sets the sweep dimensions, batch size, and snaked shape
        based on the sweeps defined in the measurement.
        """
        super().configure_from_measurement(
            setpoints + self.internal_setpoints
        )

    def get_length(self) -> int:
        """Calculates length of internal sweep"""
        return int(np.prod([len(p.get()) for p in self.internal_setpoints]))



"""Module containing abstract class for dependent readouts"""
from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
import logging
from typing import Generic, Sequence, TYPE_CHECKING, Type, TypeVar

from qm import qua

from .parameter_class import ParameterClass
from .parameters.gettable_parameter import GettableParameter
from .parameters.gettable_parameter_base import GettableParameterBase
from .parameters.gettable_parameter_multi import GettableParameterMulti
from .import path_finders
from .signal import Signal

if TYPE_CHECKING:
    from .parameters.sequence_parameter import SequenceParameter
    from .read_sequence import ReadSequence
    from qcodes.parameters import Parameter

P = TypeVar("P", bound=ParameterClass)

@dataclass(frozen = True)
class EmptyParameterClass(ParameterClass):
    pass

class AbstractReadout(Generic[P], ABC):
    """
    Abstract base class for abstract readouts. This base class handles qua
    variable and stream declaration, saving and streaming. The child class only
    needs to handle the abstract method `qua_measure`
    """
    PARAMETER_CLASS: Type[P]
    arbok_params: P

    def __init__(
        self,
        name: str,
        read_sequence: ReadSequence,
        signal: Signal,
        save_results: bool = True,
        parameters: dict | None = None
    ):
        """
        Constructor method of `AbstractReadout`
        
        Args:
            name (str): name of readout
            read_sequence (ReadSequence): Sequence generating the given readout
            signal (Signal): Signal to be used in the readout
            save_results (bool, optional): Whether to save the results
            params (dict, optional): Parameters to be added to the read sequence

        Info:
            Parameters can be added via this abstract readout to the
            read sequence. The parameters are added with the prefix of the
            readout name. For example, if the readout is called 'my_readout'
            and the parameter is called 'my_param', it will be added to the
            read sequence as 'my_readout__my_param'. Parameters are declared
            only on the read sequence and not on the abstract readout.
            The naming convention prevents name clashes with other
            readouts of the same type. You would want to declare some readout
            attributes as parameters, e.g. the voltage points in order to
            sweep and modify them in real time.
        """
        self.name = name
        self.read_sequence: ReadSequence = read_sequence
        self.signal: Signal = signal
        self.save_results = save_results
        self._parameters: dict[str, SequenceParameter] = {}
        self._gettables: dict[str, GettableParameterBase] = {}

        ### Parameters are added to the sequence with the readout prefix
        if parameters is not None:
            self.add_qc_params_from_config(parameters)
            if hasattr(self, "PARAMETER_CLASS"):
                self.arbok_params = self.map_arbok_params()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if "PARAMETER_CLASS" not in cls.__dict__:
            cls.PARAMETER_CLASS = EmptyParameterClass  # type: ignore[assignment]

    @abstractmethod
    def qua_measure(self):
        """Measures the qua variables for the given abstract readout"""
        raise NotImplementedError(
            "Abstract method 'qua_measure' not implemented in child class"
        )

    @property
    def parameters(self) -> dict[str, SequenceParameter]:
        """Returns the parameters of the readout"""
        return self._parameters

    @property
    def gettables(self) -> dict[str, GettableParameterBase]:
        """Returns the gettables of the readout"""
        return self._gettables

    def create_gettable(
        self,
        gettable_name: str,
        var_type: type[int | bool | qua.fixed]
        ) -> GettableParameter:
        """
        Creates a new GettableParameter for the AbstractReadout.
        The gettable is added to the read sequence as a parameter and registered
          under the given name.
        
        Args:
            gettable_name (str): Name of the gettable to be created
            var_type (int | bool | qua.fixed): Type of the gettable variable

        Returns:
            GettableParameter: The created gettable parameter
        """
        gettable = self.read_sequence.add_parameter(
            parameter_class = GettableParameter,
            name = gettable_name,
            read_sequence = self.read_sequence,
            var_type = var_type
        )
        self.signal.add_gettable(gettable)
        self._gettables[gettable.full_name] = gettable
        return gettable

    def create_multi_gettable(
        self,
        gettable_name: str,
        var_type: type[int | bool | qua.fixed],
        internal_setpoints: Sequence[Parameter]
    ) -> GettableParameterMulti:
        """
        Creates a new GettableParameterMulti for the AbstractReadout.
        The gettable is added to the read sequence as a parameter and registered
        under the given name.
        
        Args:
            gettable_name (str): Name of the gettable to be created
            var_type (int | bool | qua.fixed): Type of the gettable variable
            internal_setpoints (internal_setpoints: Sequence[Parameter]):
                setpoints for this multi gettable

        Returns:
            GettableParameter: The created gettable parameter
        """
        gettable = self.read_sequence.add_parameter(
            parameter_class = GettableParameterMulti,
            name = gettable_name,
            read_sequence = self.read_sequence,
            var_type = var_type,
            internal_setpoints = internal_setpoints
        )
        self.signal.add_gettable(gettable)
        self._gettables[gettable.full_name] = gettable
        return gettable

    def qua_declare_variables(self) -> None:
        """Declares all necessary qua variables for readout"""
        for gettable_name, gettable in self.gettables.items():
            logging.debug(
                "Declaring variables for gettable %s on abstract readout %s",
                gettable_name, self.name)
            gettable.qua_declare_variables()

    def qua_save_variables(self) -> None:
        """Saves the qua variables of all gettables in this readout"""
        if self.save_results:
            for gettable_name, gettable in self.gettables.items():
                logging.debug(
                    "Saving variables of gettable %s on abstract readout %s",
                    gettable_name, self.name)
                gettable.qua_save_variables()

    def qua_save_streams(self) -> None:
        """Saves acquired results to qua stream"""
        if self.save_results:
            for gettable_name, gettable in self.gettables.items():
                logging.debug(
                    "Saving streams of gettable %s on abstract readout %s",
                    gettable_name, self.name)
                gettable.qua_save_streams()
        else:
            logging.debug(
                "NOT saving streams of abstract readout %s", self.name)

    def qua_measure_and_save(self, *args, **kwargs):
        """Measures and saves the result of the given readout"""
        self.qua_measure(*args, **kwargs)
        self.qua_save_variables()

    def add_qc_params_from_config(self, param_dict: dict) -> None:
        """
        Adds the given parameters to the sequence with the readout prefix
        
        Args:
            params (list): List of parameter names to be added to the sequence
        """
        full_params = {f'{self.name}__{k}': v for k, v in param_dict.items()}
        self.read_sequence.add_qc_params_from_config(full_params)

    def map_arbok_params(self) -> P:
        """
        Adds the parameters to a ParameterClass instance

        Returns:
            ParameterClass: ParameterClass with references to all needed params
        """
        if not hasattr(self, "PARAMETER_CLASS"):
            raise AttributeError(
                f"AbstractReadout {self.name} does not have a PARAMETER_CLASS"
                " class-attribute"
            )
        arg_names = [f.name for f in fields(self.PARAMETER_CLASS) if f.init]
        arg_names_full = [
            f'{self.name}__{f}' for f in arg_names
            ]
        init_dict_full = self.read_sequence.get_parameters_and_maps(arg_names_full)
        init_dict = self.read_sequence.measurement.get_parameters_and_maps(arg_names)
        init_dict.update(self.read_sequence.get_parameters_and_maps(arg_names))
        init_dict.update(
            {n.split(f"{self.name}__")[1]: p for n, p in init_dict_full.items()}
        )
        return self.PARAMETER_CLASS(**init_dict)

    def get_gettable_from_path(self, attr_path: str) -> GettableParameter:
        """
        Returns the gettable from a given path from a given string. If the
        string leads to an AbstractReadout it is being tried to find a single
        gettable associated to that AbstractReadout
        
        Args:
            attr_path (str): Path to the given gettable relative to the
                ReadSequence with the format 'signal.gettable_name'
        
        Returns:
            gettable: The found gettable from the given path

        Raises:
            ValueError: If the path is not in the format 'signal.gettable_name'
            KeyError: If the signal or gettable is not found in the read sequence
            TypeError: If the found object is not a child class of gettable
        """
        attributes = attr_path.split('.')
        try:
            if len(attributes) == 2:
                signal_name, gettable_name = attributes
                return path_finders.get_gettable_from_read_sequence(
                    self.read_sequence, signal_name, gettable_name)
            elif len(attributes) > 2:
                return path_finders.get_gettable_from_measurement_path(
                    self.read_sequence.measurement, attributes
                )
            else:
                raise ValueError(
                    f"Path {attr_path} does not lead to a single gettable. "
                    "Please provide a path with the format 'signal.gettable_name' "
                    "or 'sub_sequence_path.signal.gettable_name'."
                )
        except AttributeError as e:
            raise ValueError(
                f"Error resolving given path: '{attr_path}' for "
                f" abstract readout {self.name}. CHECK ERROR ABOVE!"
            ) from e
